(**************************************************************************)
(*                                                                        *)
(*  Setup for lightweight web server                                      *)
(*  Copyright (C) 2024   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE SUTLS;

        (****************************************************************)
        (*                                                              *)
        (*                    PM Setup for WebServe                     *)
        (*                    Page 1 of the notebook                    *)
        (*                                                              *)
        (*        Started:        6 November 2024                       *)
        (*        Last edited:    6 November 2024                       *)
        (*        Status:         OK                                    *)
        (*                                                              *)
        (****************************************************************)


FROM SYSTEM IMPORT ADDRESS, INT16, CAST, ADR;

IMPORT OS2, OS2RTL;

IMPORT DID, Strings, CommonSettings;

FROM Languages IMPORT
    (* type *)  LangHandle,
    (* proc *)  StrToBuffer, StrToBufferN;

FROM RINIData IMPORT
    (* proc *)  RemoteOperation, INIPut, INIPutString,
                OpenINIFile, CloseINIFile, INIDeleteKey,
                INIFetch, INIGetCard, INIGetTwoShort, INIGetString,
                MakeDirectory;

FROM Remote IMPORT
    (* proc *)  OurDirectory;

FROM MiscFuncs IMPORT
    (* type *)  CharArrayPointer,
    (* proc *)  EVAL;

FROM FileOps IMPORT
    (* type *)  FilenameString;

(**************************************************************************)

CONST
    Nul = CHR(0);
    LangStringSize = 32;

TYPE
    ServiceType = (HTTP, HTTPS);
    CardArray = ARRAY ServiceType OF CARDINAL;

VAR
    INIname: ARRAY [0..31] OF CHAR;

    ChangeInProgress: BOOLEAN;
    OurPageID: CARDINAL;
    OurLang: LangHandle;
    OurLangStr: ARRAY [0..LangStringSize] OF CHAR;

    (* Handle to the window that belongs to this page.  We can save     *)
    (* this as a static value because there is never more than one      *)
    (* instance of this page.                                           *)

    pagehandle, hwndParent: OS2.HWND;
    notebookhandle: OS2.HWND;

    (* Original values of some ini variables. *)

    OldServerPort, OldTimeout: CardArray;
    OldMaxClients: CARDINAL;
    OldResolveIP: BOOLEAN;
    OldLanguage: ARRAY [0..LangStringSize] OF CHAR;

(************************************************************************)
(*                    OPERATIONS ON DIALOGUE LABELS                     *)
(************************************************************************)

PROCEDURE SetLanguage (lang: LangHandle);

    (* Relabels this page in the new language. *)

    VAR stringval: ARRAY [0..511] OF CHAR;

    BEGIN
        OurLang := lang;
        StrToBuffer (lang, "TLS.tab", stringval);
        OS2.WinSendMsg (notebookhandle, OS2.BKM_SETTABTEXT,
                        CAST(ADDRESS,OurPageID), ADR(stringval));
        StrToBuffer (lang, "TLS.enableTLS", stringval);
        OS2.WinSetDlgItemText (pagehandle, DID.EnableTLS, stringval);
        StrToBuffer (lang, "TLS.PrivateKey", stringval);
        OS2.WinSetDlgItemText (pagehandle, DID.PrivateKeyLabel, stringval);
        StrToBuffer (lang, "TLS.OurCertificate", stringval);
        OS2.WinSetDlgItemText (pagehandle, DID.OurCertificateLabel, stringval);
        StrToBuffer (lang, "TLS.CACertificate", stringval);
        OS2.WinSetDlgItemText (pagehandle, DID.CAfileLabel, stringval);
    END SetLanguage;

(************************************************************************)
(*                     LOADING DATA FROM THE INI FILE                   *)
(************************************************************************)

PROCEDURE LoadValues (hwnd: OS2.HWND);

    (* Fills the dialogue elements on page 1 with data from the INI file,       *)
    (* or loads default values if they're not in the INI file.                  *)

    VAR str: FilenameString;

    BEGIN
        EVAL (OpenINIFile (INIname));

        (* Private key file. *)

        IF NOT INIGetString ('$SYS', 'PrivateKey', str) THEN
            str := "";
        END (*IF*);
        OS2.WinSetDlgItemText (hwnd, DID.PrivateKey, str);

        (* Certificate. *)

        IF NOT INIGetString ('$SYS', 'OurCertificate', str) THEN
            str := "";
        END (*IF*);
        OS2.WinSetDlgItemText (hwnd, DID.OurCertificate, str);

        (* CA Certificate. *)

        IF NOT INIGetString ('$SYS', 'CAfile', str) THEN
            str := "";
        END (*IF*);
        OS2.WinSetDlgItemText (hwnd, DID.CAfile, str);

        CloseINIFile;

    END LoadValues;

(************************************************************************)
(*                      STORING DATA TO THE INI FILE                    *)
(************************************************************************)

(*
PROCEDURE QueryButton (hwnd: OS2.HWND;  B: CARDINAL): CARDINAL;

    BEGIN
        RETURN OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, B,
                                              OS2.BM_QUERYCHECK, NIL, NIL));
    END QueryButton;
*)

(**************************************************************************)

PROCEDURE StoreData (hwnd1: OS2.HWND);

    (* Stores the values on page 1 back into the INI file.  *)

    VAR str: FilenameString;

    BEGIN
        EVAL (OpenINIFile (INIname));

        (* Private key file. *)

        OS2.WinQueryDlgItemText (hwnd1, DID.PrivateKey, HIGH(str), str);
        INIPutString ('$SYS', 'PrivateKey', str);

        (* Certificate. *)

        OS2.WinQueryDlgItemText (hwnd1, DID.OurCertificate, HIGH(str), str);
        INIPutString ('$SYS', 'OurCertificate', str);

        (* CA Certificate. *)

        OS2.WinQueryDlgItemText (hwnd1, DID.CAfile, HIGH(str), str);
        INIPutString ('$SYS', 'CAfile', str);

        CloseINIFile;

    END StoreData;

(**************************************************************************)

PROCEDURE ["SysCall"] DialogueProc (hwnd: OS2.HWND;  msg: OS2.ULONG;
                                      mp1, mp2: OS2.MPARAM): OS2.MRESULT;

    VAR id, code: OS2.USHORT;
        value: CARDINAL;
        NewLangStr: ARRAY [0..LangStringSize] OF CHAR;

    BEGIN
        IF msg = OS2.WM_INITDLG THEN

            OS2.WinSetWindowPos (hwnd, 0, 0, 0, 0, 0, OS2.SWP_MOVE);
            LoadValues (hwnd);
            RETURN NIL;

        ELSIF msg = OS2.WM_PRESPARAMCHANGED THEN

            IF ChangeInProgress THEN
                RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
            ELSE
                ChangeInProgress := TRUE;
                CommonSettings.UpdateFontFrom (hwnd, CommonSettings.MainNotebook);
                ChangeInProgress := FALSE;
                RETURN NIL;
            END (*IF*);

        ELSIF msg = OS2.WM_CONTROL THEN

            id := OS2.USHORT1FROMMP (mp1);
            code := OS2.USHORT2FROMMP (mp1);
            value := OS2.ULONGFROMMP(mp2);
            IF code = OS2.EN_CHANGE THEN
                IF id = DID.Language THEN
                    OS2.WinQueryDlgItemText (hwnd, DID.Language,
                                             LangStringSize+1, NewLangStr);
                    IF CommonSettings.ChangeLanguageTo(NewLangStr) THEN
                        CommonSettings.CurrentLanguage (OurLang, OurLangStr);
                        SetLanguage (OurLang);
                        OS2.WinPostMsg (hwndParent, CommonSettings.LANGCHANGED,
                               OS2.MPFROMLONG(1), OS2.MPFROMLONG(0));
                    END (*IF*);
                    RETURN NIL;
                END (*IF*);
            END (*IF*);

        END (*IF*);

        RETURN OS2.WinDefDlgProc (hwnd, msg, mp1, mp2);

    END DialogueProc;

(**************************************************************************)

PROCEDURE CreatePage (notebook: OS2.HWND;
                        VAR (*OUT*) PageID: CARDINAL): OS2.HWND;

    (* Creates the TLS page and adds it to the notebook. *)

    VAR Label: ARRAY [0..31] OF CHAR;

    BEGIN
        CommonSettings.GetINIFilename (INIname);
        notebookhandle := notebook;
        hwndParent := OS2.WinQueryWindow (notebook, OS2.QW_PARENT);
        pagehandle := OS2.WinLoadDlg(notebook, notebook,
                       DialogueProc,    (* dialogue procedure *)
                       0,                   (* use resources in EXE *)
                       DID.TLSpage,          (* dialogue ID *)
                       NIL);                 (* creation parameters *)
        PageID := OS2.ULONGFROMMR (OS2.WinSendMsg (notebook, OS2.BKM_INSERTPAGE,
                         NIL, OS2.MPFROM2SHORT (OS2.BKA_MAJOR+OS2.BKA_AUTOPAGESIZE, OS2.BKA_LAST)));
        OurPageID := PageID;
        Label := "TLS";
        OS2.WinSendMsg (notebook, OS2.BKM_SETTABTEXT,
                        CAST(ADDRESS,PageID), ADR(Label));
        OS2.WinSendMsg (notebook, OS2.BKM_SETPAGEWINDOWHWND,
                        CAST(ADDRESS,PageID), CAST(ADDRESS,pagehandle));
        RETURN pagehandle;
    END CreatePage;

(**************************************************************************)

PROCEDURE SetFont (VAR (*IN*) name: CommonSettings.FontName);

    (* Sets the font of the text on this page. *)

    CONST bufsize = CommonSettings.FontNameSize;

    BEGIN
        OS2.WinSetPresParam (pagehandle, OS2.PP_FONTNAMESIZE, bufsize, name);
    END SetFont;

(**************************************************************************)

BEGIN
    INIname := "garbage.INI";
    ChangeInProgress := FALSE;
    pagehandle := OS2.NULLHANDLE;
    hwndParent := OS2.NULLHANDLE;
    OldMaxClients := 99;
    OldServerPort := CardArray{0,0};
    OldTimeout := CardArray{0,0};
    OldLanguage := "";
    OldResolveIP := FALSE;
END SUTLS.

