(**************************************************************************)
(*                                                                        *)
(*  Transport Layer Security                                              *)
(*  Copyright (C) 2024   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE TLS;

        (********************************************************)
        (*                                                      *)
        (*                Transport Layer Security              *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            2 April 2018                    *)
        (*  Last edited:        27 September 2024               *)
        (*  Status:             Still seriously incomplete      *)
        (*                                                      *)
        (********************************************************)

(************************************************************************)
(*  NOTE ABOUT VERSIONS                                                 *)
(*  Version numbering continues on from the SSL version numbers.        *)
(*  Thus we have:                                                       *)
(*          3.0 or lower    SSL versions pre-TLS                        *)
(*          3.1             TLS version 1.0                             *)
(*          3.2             TLS version 1.1                             *)
(*          3.3             TLS version 1.2                             *)
(*  The present code supports only 3.3.                                 *)
(************************************************************************)


FROM STextIO IMPORT                 (* for debugging *)
    (* proc *)  WriteChar, WriteString, WriteLn;

IMPORT Strings;

FROM SYSTEM IMPORT CAST, CARD8, CARD16, CARD32, ADR, LOC;

FROM BigNum IMPORT
    (* type *)  BN;

FROM VarStrings IMPORT
    (* type *)  ByteStringPtr;

FROM TLSBase IMPORT
    (* type *)  TextType,
    (* proc *)  NYI, WriteHex, WriteHexString, AppendHexString,
                Get3Bytes, Put3Bytes,
                Checksize, Resize;

FROM TLSHandshake IMPORT
    (* type *)  HSstate,
    (* proc *)  DoHandshake;

FROM TLSRecord IMPORT
    (* type *)  RLState,
    (* proc *)  TLSRecordInit, TLSRecordClose, FirstNBytes, GetFragment, DiscardFragment, PutFragment,
                RecordTraceOn, RecordTraceOff;

FROM TLSCipherSuites IMPORT
    (* const*)  NullSuite,
    (* type *)  CipherSuite,
    (* proc *)  NameOfSuite, AcceptCipherSuite, CodeToCipherSuite, CipherSuiteToCode;

FROM TLSCompression IMPORT
    (* type *)  CompressionMethod,
    (* proc *)  CompressionSupported;

FROM TLSCertificates IMPORT
    (* proc *)  AssembleOurCertificates, AcceptCertificates;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransaction, LogTransactionL;

FROM FileOps IMPORT
    (* type *)  ChanId,
    (* proc *)  OpenOldFile, CloseFile, ReadRaw, ReadLine;

FROM TimeConv IMPORT
    (* proc *)  time;

FROM RandCard IMPORT
    (* proc *)  RandCardinal;

FROM MiscFuncs IMPORT
    (* proc *)  AppendCard;

FROM Inet2Misc IMPORT
    (* proc *)  Swap2;

FROM Sockets IMPORT
    (* type *)  Socket,
    (* proc *)  recv, soclose;

FROM TaskControl IMPORT
    (* type *)  Lock,
    (* proc *)  CreateLock, Obtain, Release;

FROM LowLevel IMPORT
    (* proc *)  Copy, EVAL, AddOffset;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

TYPE
    Byte32 = ARRAY [0..31] OF CARD8;
    Byte46 = ARRAY [0..45] OF CARD8;    (* used? possibly not *)
    Byte48 = ARRAY [0..47] OF CARD8;    (* used only by client premaster *)

CONST
    Nul = CHR(0);
    CtrlZ = CHR(26);
    SSLversionMajor = 3;
    SSLversionMinor = 3;

    Zero32 = Byte32 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

    Zero48 = Byte48 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

    ListCipherSuites = TRUE;

    maxIDsize = 32;

TYPE
    (* The standard defines a type "opaque" for both fixed-length and   *)
    (* variable-length arrays.  In the variable-length case only, such  *)
    (* an array is passed in messages as a length followed by the data. *)
    (* The number of bytes in the length depends on the maximum array   *)
    (* length; we therefore define two types, "opaque short" and        *)
    (* "opaque medium".  These are used only for variable-length byte   *)
    (* strings.  For fixed-length, we just use a fixed-length data type.*)

    OpaqueS =   RECORD
                    length: CARD8;
                    string: ByteStringPtr;
                END (*RECORD*);

    OpaqueM =   RECORD
                    length: CARD16;
                    string: ByteStringPtr;
                END (*RECORD*);

TYPE
    sessIDtype = ARRAY [0..maxIDsize] OF CARD8;

    TLSsession = POINTER TO SessionState;

    PRFAlgorithm = (tls_prf_sha256);
    BulkCipherAlgorithm = (nocipher, rc4, tripledes, aes);
    CipherType = (stream, block, aead);
    MACAlgorithm = (null, hmac_md5, hmac_sha1, hmac_sha256, hmac_sha384,
                                 hmac_sha512);

    SecurityParameters =
            RECORD
                prf_algorithm:          PRFAlgorithm;      (* constant? *)
                bulk_cipher_algorithm:  BulkCipherAlgorithm;  (* used? *)
                cipher_type:            CipherType;        (* used? *)
                enc_key_length:         CARD8;             (* used? *)
                block_length:           CARD8;             (* used? *)
                fixed_iv_length:        CARD8;             (* used? *)
                record_iv_length:       CARD8;             (* used? *)
                mac_algorithm:          MACAlgorithm;      (* used? *)
                mac_length:             CARD8;             (* used? *)
                mac_key_length:         CARD8;             (* used? *)
                compression_algorithm:  CompressionMethod; (* set during handshake *)
                master_secret:          Byte48;            (* used? *)
                client_random:          Byte32;            (* supplied by client. used? *)
                server_random:          Byte32;            (* generated, sent to client *)
                client_premaster:       Byte48;            (* supplied by client *)

                (* Extra fields not mentioned in the standard. *)

                pending, ciphersuite:   CipherSuite;
                (*encrrule:               EncrRule;*)

            END (*RECORD*);

    SessionState =  RECORD
                        IsServer:   BOOLEAN;
                        sock:       Socket;
                        logID:      TransactionLogID;
                        sessID:     sessIDtype;
                        security:   SecurityParameters;
                        handshakestate: HSstate;
                        rlstate:    RLState;
                    END (*RECORD*);

(************************************************************************)
(*                             SESSION ID                               *)
(* For a full implementation, the client supplies a session ID during   *)
(* a handshake, and the server either accepts that ID or generates a    *)
(* new ID to indicate that it does not agree to resume the previous     *)
(* session.  In my present ID I'm not cacheing old sessions, so I       *)
(* generate a unique ID every time.                                     *)
(*                                                                      *)
(* NOTE: I suspect that this is not a TLS 1.3 feature.                  *)
(************************************************************************)

VAR NextID: sessIDtype;
    NextIDLock: Lock;

(************************************************************************)

PROCEDURE ResetNextID;

    VAR j: CARDINAL;

    BEGIN
        Obtain (NextIDLock);
        NextID[0] := maxIDsize;
        FOR j := 1 TO maxIDsize DO
            NextID[j] := 0;
        END (*FOR*);
        Release (NextIDLock);
    END ResetNextID;

(************************************************************************)
(*                      OPENING AND CLOSING A SESSION                   *)
(************************************************************************)

PROCEDURE CreateSessionRecord (S: Socket;  as_server: BOOLEAN;
                                    logID: TransactionLogID): TLSsession;

    (* Sets up the initial TLS state.    *)

    VAR result: TLSsession;

    BEGIN
        NEW (result);
        result^.IsServer := as_server;
        result^.sock := S;
        result^.logID := logID;
        WITH result^.security DO
            prf_algorithm           := tls_prf_sha256;
            bulk_cipher_algorithm   := nocipher;
            cipher_type             := stream;
            enc_key_length          := 0;
            block_length            := 0;
            fixed_iv_length         := 0;
            record_iv_length        := 0;
            mac_algorithm           := null;
            mac_length              := 0;
            mac_key_length          := 0;
            compression_algorithm   := nocompression;
            master_secret           := Zero48;
            client_random           := Zero32;
            server_random           := Zero32;
            pending                 := NullSuite;
            ciphersuite             := NullSuite;
        END (*WITH*);
        RETURN result;
    END CreateSessionRecord;

(************************************************************************)

PROCEDURE OpenTLSsession (S: Socket;  as_server: BOOLEAN;
                           logID: TransactionLogID;
                            VAR (*OUT*) session: TLSsession): BOOLEAN;

    (* Opens a new connection.  Returns FALSE if the operation failed.  *)

    VAR handshakestate: HSstate;

    BEGIN
        session := CreateSessionRecord (S, as_server, logID);
        session^.rlstate := TLSRecordInit(S, as_server, logID);
        handshakestate := DoHandshake (S, as_server, session^.rlstate, logID);
        IF handshakestate = CAST(HSstate,NIL) THEN
            CloseTLSsession (session);
            RETURN FALSE
        ELSE
            session^.handshakestate := handshakestate;
            RETURN TRUE;
        END (*IF*);
    END OpenTLSsession;

(************************************************************************)

PROCEDURE CloseTLSsession (VAR (*INOUT*) sess: TLSsession);

    (* Closes the session, deletes the session data. *)

    BEGIN
        (* There are probably some formalities I have to do here, but   *)
        (* I'll come back to that later.                                *)

        IF sess <> NIL THEN
            TLSRecordClose (sess^.rlstate);
            soclose (sess^.sock);
            DISPOSE (sess);
        END (*IF*);
    END CloseTLSsession;

(************************************************************************)
(*                    SENDING DATA TO RECORD LAYER                      *)
(************************************************************************)

PROCEDURE TLSsend (sess: TLSsession;  VAR (*IN*) message: ARRAY OF LOC;
                                                length: CARDINAL): CARDINAL;

    (* Sends a message of "length" bytes.  The returned value is the    *)
    (* number of bytes sent, or MAX(CARDINAL) if there was an error.    *)

    BEGIN
        (* NOT YET IMPLEMENTED *)
        RETURN MAX(CARDINAL);
    END TLSsend;

(************************************************************************)

PROCEDURE TLSrecv (sess: TLSsession;  VAR (*OUT*) message: ARRAY OF LOC;
                          length: CARDINAL;  VAR (*OUT*) actual: CARDINAL);

        (* NOT YET IMPLEMENTED *)

    (* Receives a message of up to "length" bytes.  The "actual"        *)
    (* parameter returns the actual number of bytes received, or        *)
    (* MAX(CARDINAL) if there was an error.                             *)

    BEGIN
        (* NOT YET IMPLEMENTED *)
    END TLSrecv;

(************************************************************************)

PROCEDURE TLSrecvLine (sess: TLSsession;  VAR (*OUT*) message: ARRAY OF LOC;
                          length: CARDINAL;  VAR (*OUT*) actual: CARDINAL);

        (* NOT YET IMPLEMENTED *)

    (* Like TLSrecv, but returns data up to the next CRLF line          *)
    (* terminator.  The CRLF is discarded.                              *)

    BEGIN
        (* NOT YET IMPLEMENTED *)
    END TLSrecvLine;

(************************************************************************)

BEGIN
    CreateLock (NextIDLock);
    ResetNextID;
END TLS.

