(**************************************************************************)
(*                                                                        *)
(*  Transport Layer Security                                              *)
(*  Copyright (C) 2024   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE TLSCertificates;

        (********************************************************)
        (*                                                      *)
        (*                   TLS certificates                   *)
        (*      For now I'm handling only RSA certificates      *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            5 September 2018                *)
        (*  Last edited:        9 October 2024                  *)
        (*  Status:             Working so far                  *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT LOC, CARD8, ADR;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn;

FROM ASN1 IMPORT
    (* proc *)  ExtractParts, INTEGERtoBN, BitStringToBN, GetObjID;

IMPORT Strings, BigNum;

FROM INIData IMPORT
    (* type *)  HINI,
    (* proc *)  ChooseDefaultINI, OpenINIFile, INIValid, CloseINIFile, INIGetString;

FROM VarStrings IMPORT
    (* type *)  ByteStr, ByteStringPtr,
    (* proc *)  MakeBS, BSExpand, TrimBS;

FROM TLSBase IMPORT
    (* proc *)  NYI, Get3Bytes, Put3Bytes, Checksize, Resize, AppendHexString;

FROM BigNum IMPORT
    (* type *)  BN;

FROM MiscFuncs IMPORT
    (* proc *)  AppendCard;

FROM Base64 IMPORT
    (* proc *)  Decode;

FROM FileOps IMPORT
    (* const*)  NoSuchChannel,
    (* type *)  ChanId, FilenameString,
    (* proc *)  OpenOldFile, OpenNewFile, CloseFile, ReadRaw, WriteRaw, ReadLine,
                Exists, DeleteFile, MoveFile;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransaction;

FROM LowLevel IMPORT
    (* proc *)  EVAL, Copy, AddOffset;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)
(* In this module we represent a certificate or a certificate chain by  *)
(* a byte string and the length of that string.  The first ten bytes    *)
(* are overhead, as follows:                                            *)
(*      (a) first 4 bytes   reserved for handshake header               *)
(*      (b) next 3 bytes    total remaining length                      *)
(*      (c) next 3 bytes    length of the certificate itself            *)
(* In the present implementation the certificate chain is only one      *)
(* certificate long, therefore (b) is just (c) plus 3, and the string   *)
(* length is (b) plus 7.  We accept this redundancy since the caller    *)
(* needs only know about the total length.  Also, (b) will change once  *)
(* we allow chains of more than one certificate.                        *)
(************************************************************************)

CONST
    CR = CHR(13);  LF = CHR(10);  CtrlZ = CHR(26);
    CertDir = "C:\etc\pki\tls\private";
    CADir = "C:\etc\pki\tls\private\cert2";

VAR
    OurCertificate, CACertificate: FilenameString;

(* From RFC 3280, page 15. *)
(*
The X.509 v3 certificate basic syntax is as follows. For signature
calculation, the data that is to be signed is encoded using the ASN.1
distinguished encoding rules (DER) [X.690]. ASN.1 DER encoding is a
tag, length, value encoding system for each element. *)

(*
Certificate ::= SEQUENCE {
                         tbsCertificate: TBSCertificate,
                         signatureAlgorithm: AlgorithmIdentifier,
                         signatureValue: BIT STRING }
TBSCertificate ::= SEQUENCE {
                            version: [0] EXPLICIT Version DEFAULT v1,
                            serialNumber: CertificateSerialNumber,
                            signature : AlgorithmIdentifier,
                            issuer: Name,
                            validity: Validity,
                            subject: Name,
                            subjectPublicKeyInfo: SubjectPublicKeyInfo,
                            issuerUniqueID: [1] IMPLICIT UniqueIdentifier OPTIONAL,
                            -- If present, version MUST be v2 or v3
                            subjectUniqueID: [2] IMPLICIT UniqueIdentifier OPTIONAL,
                            -- If present, version MUST be v2 or v3
                            extensions: [3] EXPLICIT Extensions OPTIONAL
                            -- If present, version MUST be v3
                            }
Version ::= INTEGER {v1(0), v2(1), v3(2)}
CertificateSerialNumber ::= INTEGER
Validity ::= SEQUENCE {
                      notBefore: Time,
                      notAfter: Time }
Time ::= CHOICE {
                utcTime: UTCTime,
                generalTime: GeneralizedTime }
UniqueIdentifier ::= BIT STRING
SubjectPublicKeyInfo ::= SEQUENCE {
                                  algorithm: AlgorithmIdentifier,
                                  subjectPublicKey: BIT STRING }
Extensions ::= SEQUENCE SIZE (1..MAX) OF Extension
Extension ::= SEQUENCE {
                       extnID: OBJECT IDENTIFIER,
                       critical: BOOLEAN DEFAULT FALSE,
                       extnValue: OCTET STRING
                       }
*)

(************************************************************************)
(*                              TEST CODE                               *)
(************************************************************************)

PROCEDURE DumpASN (N: CARDINAL;  length: CARDINAL;  p: ByteStringPtr);

    VAR cid: ChanId;
        fname: ARRAY [0..15] OF CHAR;

    BEGIN
        fname := "ASNpart";
        fname[7] := CHR(N+ORD('0'));
        fname[8] := CHR(0);
        Strings.Append (".bin", fname);
        IF Exists (fname) THEN
            DeleteFile(fname);
        END (*IF*);
        cid := OpenNewFile (fname, TRUE);
        WriteRaw (cid, p^, length);
        CloseFile (cid);
    END DumpASN;

(************************************************************************)
(*                               OBJECT ID                              *)
(************************************************************************)

TYPE ID7 = ARRAY [0..6] OF CARDINAL;

CONST sha256_RSA = ID7 {1, 2, 840, 113549, 1, 1, 11};

(************************************************************************)

PROCEDURE MatchNamek (k: CARDINAL;  ID, candidate: ARRAY OF CARDINAL): BOOLEAN;

    (* Matches an object ID of length k. *)

    VAR j: CARDINAL;

    BEGIN
        j := 0;
        LOOP
            IF j >= k THEN RETURN TRUE;
            ELSIF ID[j] <> candidate[j] THEN RETURN FALSE;
            ELSE INC(j);
            END (*IF*);
        END (*LOOP*);
    END MatchNamek;

(************************************************************************)

PROCEDURE GetObjectName (VAR (*ID*) ID: ARRAY OF CARDINAL;
                            length: CARDINAL;
                              VAR (*OUT*) type: ARRAY OF CARD8;
                              VAR (*OUT*) name: ARRAY OF CHAR);

    (* Identifies an object ID. Since there are many possible ID codes, *)
    (* in this version I'm only identifying a subset.                   *)

    VAR success: BOOLEAN;

    BEGIN
        success := FALSE;
        IF length = 7 THEN
            IF MatchNamek (7, ID, sha256_RSA) THEN
                Strings.Assign ("sha256_RSA", name);
                type[0] := 4;  type[1] := 1;
                success := TRUE;
            END (*IF*);
        END (*IF*);
        IF NOT success THEN
            Strings.Assign ("unknown", name);
            type[0] := 0;  type[1] := 0;
        END (*IF*);
    END GetObjectName;

(************************************************************************)
(*                       RECEIVING A CERTIFICATE                        *)
(************************************************************************)

PROCEDURE AcceptCertificate (logID: TransactionLogID;
                        pdata: ByteStringPtr;  offset, length: CARDINAL;
                        VAR (*OUT*) modulus, exponent: BN);

    (* Accepts a certificate sent by the peer.  The returned modulus    *)
    (* and exponent are the public key.                                 *)

    CONST BNbufsize = 600;

    VAR message: ARRAY [0..800] OF CHAR;
        BNbuf: ARRAY [0..BNbufsize-1] OF CHAR;
        keybuff: ARRAY [0..4095] OF CARD8;
        part: ARRAY [0..31] OF ByteStr;
        pcert, pk: ByteStringPtr;
        j, N: CARDINAL;
        V: BN;
        values: ARRAY [0..15] OF CARDINAL;

    BEGIN
        length := Get3Bytes (pdata^, offset);
        message := "Length of certificate is ";
        AppendCard (length, message);
        LogTransaction (logID, message);

        (*
        message := "First few bytes of the certificate are ";
        AppendHexString (pdata^, offset, 16, TRUE, message);
        LogTransaction (logID, message);
        *)
        pcert := AddOffset (pdata, offset);
        N := ExtractParts (pcert^, part);

        (* There should be three parts. For testing, dump those 3 parts.*)

        IF N = 3 THEN
            FOR j := 0 TO N-1 DO
                DumpASN (j+1, part[j].size, part[j].data);
            END (*FOR*);
        ELSE
            message := "certificate has ";
            AppendCard (N, message);
            Strings.Append (" parts.", message);
            LogTransaction (logID, message);
        END (*IF*);

        (* Part 2 is signature value. *)

        (*
        message := "Signature is ";
        V := BitStringToBN (part[2].val^, 0);
        BigNum.ToHex (V, BNbuf, BNbufsize);
        Strings.Append (BNbuf, message);
        LogTransaction (logID, message);
        *)

        (* Part 0 is the certificate itself.  Break that into parts.  *)

        N := ExtractParts (part[0].data^, part);
        message := "Body of certificate has ";
        AppendCard (N, message);
        Strings.Append (" parts.", message);
        LogTransaction (logID, message);

        (* For now I'm only interested in part 6, which is the public   *)
        (* key.  It has further substructure, as shown below.           *)

        EVAL (ExtractParts (part[6].data^, part));
        pk := part[1].data;         (* Save part 1 for later. *)

        (* part[0] contains the object identifier for the encryption    *)
        (* method.  For now I'm assuming RSA encryption, whose object   *)
        (* ID is {1 2 840 113549 1 1 1}.  In a future version I'll have *)
        (* to check this.                                               *)

        EVAL (ExtractParts (part[0].data^, part));
        N := GetObjID (part[0].data^, 0, values);
        message := "Encryption = ";
        IF N = 0 THEN
            Strings.Append ("<missing>", message);
        ELSE
            Strings.Append ("{", message);
            FOR j := 0 TO N-1 DO
                IF j > 0 THEN
                    Strings.Append (" ", message);
                END (*IF*);
                AppendCard (values[j], message);
            END (*FOR*);
            Strings.Append ("}", message);
        END (*IF*);
        LogTransaction (logID, message);

        (* pk^ is the public key, which is itself an ASN.1 sequence of  *)
        (* modulus and exponent.  So we have one extra step of ASN.1    *)
        (* decoding.                                                    *)

        V := BitStringToBN (pk^, 0);
        (*
        message := "Public key = ";
        BigNum.ToHex (V, BNbuf, BNbufsize);
        Strings.Append (BNbuf, message);
        LogTransaction (logID, message);
        *)

        EVAL (BigNum.BNtoBytes (V, keybuff));
        EVAL (ExtractParts (keybuff, part));

        (*
        message := "Public key modulus = ";
        modulus := INTEGERtoBN (part[0].val^, 0);
        BigNum.ToHex (modulus, BNbuf, BNbufsize);
        Strings.Append (BNbuf, message);
        LogTransaction (logID, message);
        *)

        message := "Public key exponent = ";
        exponent := INTEGERtoBN (part[1].data^, 0);
        BigNum.ToHex (exponent, BNbuf, BNbufsize);
        Strings.Append (BNbuf, message);
        LogTransaction (logID, message);

    END AcceptCertificate;

(************************************************************************)

PROCEDURE AcceptCertificates (logID: TransactionLogID;
                        pdata: ByteStringPtr;  offset, length: CARDINAL;
                        VAR (*OUT*) modulus, exponent: BN): BOOLEAN;

    (* Accepts a chain of certificates sent by the peer.  (But, in this *)
    (* version, we only look at the first one.)  The returned modulus   *)
    (* and exponent are the public key.                                 *)

    (*VAR message: ARRAY [0..127] OF CHAR;*)

    BEGIN
        (*
        message := "In AcceptCertificates, first few bytes are ";
        AppendHexString (pdata^, offset, 16, TRUE, message);
        LogTransaction (logID, message);
        *)
        length := Get3Bytes (pdata^, offset);
        (*
        message := "Length of certificates is ";
        AppendCard (length, message);
        LogTransaction (logID, message);
        *)
        AcceptCertificate (logID, pdata, offset, length, modulus, exponent);
        RETURN TRUE;
    END AcceptCertificates;

(************************************************************************)
(*                      OPENING A CERTIFICATE FILE                       *)
(************************************************************************)

PROCEDURE ReadOurCertificate (overhead: CARDINAL;
                                VAR (*OUT*) filename: ARRAY OF CHAR;
                                  VAR (*OUT*) length: CARDINAL;
                                    VAR (*OUT*) pdata: ByteStringPtr);

    (* overhead is the number of bytes that must be left at the start   *)
    (* of pdata^, to be filled in by the caller.  If the certificate    *)
    (* does not exist, or cannot be read, we return with length=0 and   *)
    (* pdata = NIL.                                                     *)

    CONST
        relname = "testcert.ca.crt";
        binname = "testcert.ca.crt";
        binary = FALSE;     (* setting this TRUE gives "bad certificate" error *)
        InitialAllocation = 256;

    VAR cid: ChanId;
        allocated, nread: CARDINAL;
        line: ARRAY [0..511] OF CHAR;

    BEGIN
        allocated := InitialAllocation;
        ALLOCATE (pdata, allocated);
        length := overhead; (* Filling in those initial bytes, if any,  *)
                            (* is the caller's problem.                 *)

        (* Create a file name. *)

        Strings.Assign (CertDir, filename);
        Strings.Append ("\", filename);
        IF binary THEN
            Strings.Append (binname, filename);
        ELSE
            Strings.Append (relname, filename);
        END (*IF*);

        cid := OpenOldFile (filename, FALSE, binary);
        LOOP
            IF binary THEN
                ReadRaw (cid, line, 512, nread);
            ELSE
                REPEAT
                    ReadLine (cid, line);
                UNTIL (line[0] <> '-') OR (line[1] <> '-') OR (line[2] <> '-');
                IF line[0] = CtrlZ THEN
                    nread := 0;
                ELSE
                    nread := Strings.Length (line);
                END (*IF*);
            END (*IF*);
            IF nread = 0 THEN
                EXIT (*LOOP*);
            ELSE
                Checksize (pdata, allocated, length, nread);
                Copy (ADR(line), ADR(pdata^[length]), nread);
                INC (length, nread);
            END (*IF*);
        END (*LOOP*);
        CloseFile (cid);
        IF length = overhead THEN
            (* No data were read. *)
            DEALLOCATE (pdata, allocated);
            length := 0;
        ELSE
            Resize (pdata, allocated, length);
        END (*IF*);

    END ReadOurCertificate;

(************************************************************************)
(*               GETTING DATA FROM OUR OWN CERTIFICATE                  *)
(************************************************************************)

PROCEDURE GetSignatureType (VAR (*OUT*) type: ARRAY OF CARD8);

    (* The two-byte result specifies a hash algorithm and a signature   *)
    (* algorithm, specifying what sort of signature our own certificate *)
    (* has.  If there is no certificate, or if the certificate does not *)
    (* have a signature, we return (0,0).                               *)

    VAR length, N: CARDINAL;
        pdata: ByteStringPtr;
        part: ARRAY [0..31] OF ByteStr;
        objID: ARRAY [0..31] OF CARDINAL;
        filename, line: ARRAY [0..511] OF CHAR;

    BEGIN
        ReadOurCertificate (0, filename, length, pdata);
        IF length = 0 THEN
            type[0] := 0;  type[1] := 0;
        ELSE
            N := ExtractParts (pdata^, part);
            IF N < 2 THEN
                type[0] := 0;  type[1] := 0;
            ELSE
                (* Signature algorithm is part 1. *)

                EVAL (ExtractParts (part[1].data^, part));
                N := GetObjID (part[0].data^, 0, objID);
                GetObjectName (objID, N, type, line);
            END (*IF*);
            DEALLOCATE (pdata, length);
        END (*IF*);

        (*  The rest is only trace code. *)

        (*
        line := "Certificate signature algorithm is (";
        AppendCard (type[0], line);
        Strings.Append (',', line);
        AppendCard (type[1], line);
        Strings.Append (')', line);
        WriteString (line);  WriteLn;
        *)
    END GetSignatureType;

(************************************************************************)
(*              PREPARING OUR CERTIFICATES FOR TRANSMISSION             *)
(************************************************************************)

PROCEDURE AssembleOneCertificate (ID: TransactionLogID;
                                    VAR (*IN*) filename: ARRAY OF CHAR;
                                    VAR (*OUT*) result: ByteStr;
                                    VAR (*INOUT*) pos: CARDINAL): CARDINAL;

    (* At position pdata^[pos]pos of the result string, stores a three-byte length *)
    (* followed by the file content, and updates pos.  Expands result   *)
    (* if necessary.  Returns the length (file size + 3).  *)
    (* If the certificate    *)
    (* does not exist, or cannot be read, we return with length=0 and   *)
    (* pdata = NIL.                                                     *)

    CONST binary = TRUE;
        linesize = 1500;

    VAR cid: ChanId;
        nread, length, pos0: CARDINAL;
        q: ByteStringPtr;
        line: ARRAY [0..linesize-1] OF CHAR;
        message: ARRAY [0..255] OF CHAR;

    BEGIN
        pos0 := pos;
        BSExpand (result, 3);
        INC (result.size, 3);
        INC (pos, 3);
        length := 0;

        cid := OpenOldFile (filename, FALSE, binary);
        REPEAT
            ReadRaw (cid, line, linesize, nread);
            IF nread > 0 THEN
                BSExpand (result, nread);
                Copy (ADR(line), ADR(result.data^[pos]), nread);
                INC (result.size, nread);
                INC (length, nread);
                INC (pos, nread);
            END (*IF*);
        UNTIL nread = 0;
        CloseFile (cid);
        q := ADR(result.data^[pos0]);
        Put3Bytes (q^, length);

        message := "Assembled certificate ";  Strings.Append (filename, message);
        Strings.Append (", length ", message);  AppendCard (length, message);
        LogTransaction (ID, message);

        INC (length, 3);

        RETURN length;

    END AssembleOneCertificate;

(************************************************************************)

PROCEDURE AssembleOurCertificates (VAR (*OUT*) length: CARDINAL;
                                    VAR (*OUT*) pdata: ByteStringPtr;
                                    ID: TransactionLogID);

    (* Prepares the certificate data for transmission.  In this version *)
    (* I use a test certificate with a fixed filename, and there is     *)
    (* only one certificate in the chain.                               *)

    (* The required format is a 3-byte total length; then, for each     *)
    (* certificate, a 3-byte length followed by the certificate itself. *)
    (* These are preceded by the 4-byte overhead required by the        *)
    (* handshake protocol.                                              *)

    (* That suggests that I need to define a new data type that allows  *)
    (* linking of nested structures, but for now I can't think of an    *)
    (* elegant way to do this.                                          *)

    (* This is the version that sends two certificates. *)

    VAR q: ByteStringPtr;
        result: ByteStr;
        pos: CARDINAL;

    BEGIN
        MakeBS (result, 7);
        pos := 7;

        (* First file. *)

        length := AssembleOneCertificate (ID, OurCertificate, result, pos);

        (* Second file. *)

        INC (length, (AssembleOneCertificate (ID, CACertificate, result, pos)));

        q := ADR (result.data^[4]);
        Put3Bytes (q^, length);
        length := result.size;
        pdata := result.data;
        TrimBS (result);

    END AssembleOurCertificates;

(************************************************************************)
(*                      SETTING THE CERTIFICATE NAMES                   *)
(************************************************************************)

PROCEDURE LoadFile (filename: ARRAY OF CHAR;
                            VAR (*OUT*) buffer: ARRAY OF CARD8): CARDINAL;

    (* Loads a byte string from a file, returns the byte count. *)

    CONST BlockSize = 4096;

    VAR N, NumberRead: CARDINAL;
        cid: ChanId;
        p: POINTER TO ARRAY [0..MAX(CARDINAL) DIV 4] OF CARD8;

    BEGIN
        p := ADR(buffer);  N := 0;
        cid := OpenOldFile (filename, FALSE, TRUE);
        LOOP
            ReadRaw (cid, p^, BlockSize, NumberRead);
            IF NumberRead = 0 THEN
                EXIT (*LOOP*);
            ELSE
                INC (N, NumberRead);
                p := AddOffset (p, NumberRead);
            END (*IF*);
        END (*LOOP*);
        CloseFile (cid);
        RETURN N;
    END LoadFile;

(************************************************************************)

PROCEDURE ConvertToBinary (VAR (*INOUT*) buffer: ARRAY OF CARD8;
                                VAR (*IN*) binfile: ARRAY OF CHAR);

    (* Converts a Base64 string to binary.  The string probably extends *)
    (* over several lines.                                              *)

    (* Remark: since this will compress the result, we can afford to    *)
    (* overwrite the input in producing the output.                     *)

    CONST CR = 13; LF = 10;

    VAR j: CARDINAL;

    (********************************************************************)

    PROCEDURE FindEOL;

        (* Advances j to the next carriage return. *)

        BEGIN
            WHILE buffer[j] <> CR DO
                INC (j);
            END (*WHILE*);
        END FindEOL;

    (********************************************************************)

    VAR txtbuf: ARRAY [0..511] OF CHAR;
        p: ByteStringPtr;
        j0, amount, total: CARDINAL;
        cid: ChanId;

    BEGIN
        (* Skip past the initial line. *)

        j := 0;
        FindEOL;
        INC (j);
        IF buffer[j] = LF THEN
            INC(j);
        END(*IF*);
        total := 0;

        (* Now the main translation. *)

        p := ADR(buffer);
        WHILE buffer[j] <> ORD('-') DO
            j0 := j;
            FindEOL;
            IF j > j0 THEN
                Copy (ADR(buffer[j0]), ADR(txtbuf), j-j0);
                amount := Decode (txtbuf, p^);
                p := AddOffset (p, amount);
                INC (total, amount);
                INC (j);
                IF buffer[j] = LF THEN INC(j) END(*IF*);
            END (*IF*);
        END (*WHILE*);

        (* Write out the binary file. *)

        cid := OpenNewFile (binfile, TRUE);
        WriteRaw (cid, buffer, total);
        CloseFile (cid);

    END ConvertToBinary;

(************************************************************************)

PROCEDURE CheckForBinary (VAR (*INOUT*) file: ARRAY OF CHAR);

    (* If necessary, converts certificate file from Base64 encoded PEM  *)
    (* to a binary ASN.1 file.                                          *)

    CONST InBufSize = 65536;

    VAR cid: ChanId;
        buffer: ARRAY [0..0] OF CHAR;
        found: BOOLEAN;
        nread, pos, N: CARDINAL;
        basename, binfile, BAKname: FilenameString;
        inbuf: ARRAY [0..InBufSize-1] OF CARD8;

    BEGIN
        cid := OpenOldFile (file, FALSE, TRUE);
        IF cid <> NoSuchChannel THEN
            ReadRaw (cid, buffer, 1, nread);
            IF (nread <> 0) AND (buffer[0] = '-') THEN

                (* We have a PEM encoded file. *)

                Strings.Assign (file, basename);
                Strings.FindPrev ('.', basename, LENGTH(basename) - 1,
                                                             found, pos);
                IF found THEN
                    basename[pos] := CHR(0);
                END (*IF*);
                binfile := basename;
                Strings.Append (".bin", binfile);
                BAKname := basename;
                Strings.Append (".BAK", BAKname);
        
                (* Do the translation. *)
        
                N := LoadFile (file, inbuf);
                IF N > 0 THEN
                    DeleteFile (BAKname);
                    IF Exists (binfile) THEN
                        EVAL (MoveFile (binfile, BAKname));
                    END (*IF*);
                    ConvertToBinary (inbuf, binfile);
                    Strings.Assign (binfile, file);
                END (*IF*);
            END (*IF*);
        END (*IF*);
    END CheckForBinary;

(************************************************************************)

PROCEDURE GetCertificateNames (ININame: ARRAY OF CHAR);

    (* Loads certificate names from the INI file. *)

    VAR hini: HINI;
        SYSapp: ARRAY [0..7] OF CHAR;

    BEGIN
        OurCertificate := "";
        CACertificate := "";
        SYSapp := '$SYS';
        hini := OpenINIFile(ININame);
        IF INIValid (hini) THEN
            EVAL (INIGetString (hini, SYSapp, 'OurCertificate', OurCertificate));
            EVAL (INIGetString (hini, SYSapp, 'CAfile', CACertificate));
            CloseINIFile (hini);
        END (*IF*);
        CheckForBinary (OurCertificate);
        CheckForBinary (CACertificate);
    END GetCertificateNames;

(************************************************************************)

END TLSCertificates.

