(**************************************************************************)
(*                                                                        *)
(*  Transport Layer Security                                              *)
(*  Copyright (C) 2024   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE TLSHandshake;

        (********************************************************)
        (*                                                      *)
        (*                TLS Handshake Protocol                *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            3 July 2023                     *)
        (*  Last edited:        13 November 2024                *)
        (*  Status:             Recovering after major loss     *)
        (*                                                      *)
        (********************************************************)

(************************************************************************)
(*  NOTE ABOUT VERSIONS                                                 *)
(*  Version numbering continues on from the SSL version numbers.        *)
(*  Thus we have:                                                       *)
(*          3.0 or lower    SSL versions pre-TLS                        *)
(*          3.1             TSL version 1.0                             *)
(*          3.2             TSL version 1.1                             *)
(*          3.3             TSL version 1.2                             *)
(*  The present code supports only 3.3.                                 *)
(************************************************************************)


FROM STextIO IMPORT                 (* for debugging *)
    (* proc *)  WriteChar, WriteString, WriteLn;

IMPORT Strings, TLSRSA, RSAKeys, BigNum, RSA;

FROM SYSTEM IMPORT CAST, CARD8, CARD16, CARD32, ADDRESS, ADR, LOC;

FROM BigNum IMPORT
    (* type *)  BN;

FROM VarStrings IMPORT
    (* type *)  ByteStringPtr, ByteStr,
    (* proc *)  MakeBS;

FROM TLSBase IMPORT
    (* type *)  TextType,
    (* proc *)  NYI, WriteHex, AppendHex, AppendHex4, WriteHexString, AppendHexString,
                Get3Bytes, Put3Bytes,
                Checksize, Resize;

FROM TLSRecord IMPORT
    (* type *)  RLState,
    (* proc *)  TLSRecordInit, TLSRecordClose, FirstNBytes, GetFragment, DiscardFragment, PutFragment,
                RecordTraceOn, RecordTraceOff;

FROM TLSPRF IMPORT
    (* type *)  PRFAlgorithm,
    (* proc *)  PRF;

FROM TLSCipherSuites IMPORT
    (* const*)  NullSuite,
    (* type *)  CipherSuite,
    (* proc *)  NameOfSuite, AcceptCipherSuite, CodeToCipherSuite, CipherSuiteToCode,
                KeyBlockSize, Commit;

FROM TLSCompression IMPORT
    (* type *)  CompressionMethod,
    (* proc *)  CompressionSupported;

FROM TLSCertificates IMPORT
    (* proc *)  GetCertificateNames, AssembleOurCertificates, AcceptCertificates;

FROM SHA2 IMPORT
    (* type *)  SHA2_CTX,
    (* proc *)  SHA256Init, SHA2Update, SHA2Final;

FROM INIData IMPORT
    (* type *)  HINI,
    (* proc *)  OpenINIFile, INIValid, CloseINIFile, INIGetString;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransaction, LogTransactionL;

FROM FileOps IMPORT
    (* type *)  ChanId, FilenameString,
    (* proc *)  OpenOldFile, CloseFile, ReadRaw, ReadLine;

FROM TimeConv IMPORT
    (* proc *)  time;

FROM RandCard IMPORT
    (* proc *)  RandCardinal;

FROM MiscFuncs IMPORT
    (* proc *)  AppendCard;

FROM Inet2Misc IMPORT
    (* proc *)  Swap2;

FROM Sockets IMPORT
    (* type *)  Socket,
    (* proc *)  recv, soclose;

FROM TaskControl IMPORT
    (* type *)  Lock,
    (* proc *)  CreateLock, Obtain, Release;

FROM LowLevel IMPORT
    (* proc *)  Copy, EVAL, AddOffset;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

TYPE
    Byte32 = ARRAY [0..31] OF CARD8;
    Byte46 = ARRAY [0..45] OF CARD8;    (* used? possibly not *)
    Byte48 = ARRAY [0..47] OF CARD8;    (* used only by client premaster *)

    HSstate = POINTER TO SessionState;

    (********************************************************************)
    (* NOTE: At present I am doubling up information, with a session    *)
    (* state that has the same fields as in module TLS.  In time I hope *)
    (* to clear up the issue of which data are needed by which module.  *)
    (********************************************************************)

CONST
    Nul = CHR(0);
    CtrlZ = CHR(26);
    SSLversionMajor = 3;
    SSLversionMinor = 3;

    Zero32 = Byte32 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

    Zero48 = Byte48 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

    ListCipherSuites = TRUE;

    maxIDsize = 32;

TYPE
    (* The standard defines a type "opaque" for both fixed-length and   *)
    (* variable-length arrays.  In the variable-length case only, such  *)
    (* an array is passed in messages as a length followed by the data. *)
    (* The number of bytes in the length depends on the maximum array   *)
    (* length; we therefore define two types, "opaque short" and        *)
    (* "opaque medium".  These are used only for variable-length byte   *)
    (* strings.  For fixed-length, we just use a fixed-length data type.*)

    OpaqueS =   RECORD
                    length: CARD8;
                    string: ByteStringPtr;
                END (*RECORD*);

    OpaqueM =   RECORD
                    length: CARD16;
                    string: ByteStringPtr;
                END (*RECORD*);

TYPE
    sessIDtype = ARRAY [0..maxIDsize] OF CARD8;

    BulkCipherAlgorithm = (nocipher, rc4, tripledes, aes);
    CipherType = (stream, block, aead);
    MACAlgorithm = (null, hmac_md5, hmac_sha1, hmac_sha256, hmac_sha384,
                                 hmac_sha512);

    SecurityParameters =
            RECORD
                prf_algorithm:          PRFAlgorithm;      (* constant? *)
                bulk_cipher_algorithm:  BulkCipherAlgorithm;  (* used? *)
                cipher_type:            CipherType;        (* used? *)
                enc_key_length:         CARD8;             (* used? *)
                block_length:           CARD8;             (* used? *)
                fixed_iv_length:        CARD8;             (* used? *)
                record_iv_length:       CARD8;             (* used? *)
                mac_algorithm:          MACAlgorithm;      (* used? *)
                mac_length:             CARD8;             (* used? *)
                mac_key_length:         CARD8;             (* used? *)
                compression_algorithm:  CompressionMethod; (* set during handshake *)
                master_secret:          Byte48;            (* calculated and used *)
                client_random:          Byte32;            (* supplied by client. used *)
                server_random:          Byte32;            (* generated, sent to client *)
                client_premaster:       Byte48;            (* supplied by client *)

                (* Extra fields not mentioned in the standard. *)

                pending, ciphersuite:   CipherSuite;
                (*encrrule:               EncrRule;*)
                keyblock:               ARRAY [0..255] OF CARD8;   (* used in this module *)

            END (*RECORD*);

    SessionState =  RECORD
                        IsServer:   BOOLEAN;
                        sock:       Socket;
                        logID:      TransactionLogID;
                        sessID:     sessIDtype;
                        security:   SecurityParameters;
                        rlstate:    RLState;
                        hashctx1,
                         hashctx2:  SHA2_CTX;
                    END (*RECORD*);

VAR PrivateKeyFile: FilenameString;

(************************************************************************)
(*                              DEBUGGING STUFF                         *)
(************************************************************************)

(*
PROCEDURE WriteAscString (VAR (*IN*) buf: ARRAY OF CARD8;  pos, N: CARDINAL);

    (* Like WriteHexString, but we interpret the bytes as characters.  *)

    VAR j: CARDINAL;

    BEGIN
        FOR j := 0 TO N-1 DO
            WriteChar (CHR(buf[pos]));  INC(pos);
        END (*FOR*);
    END WriteAscString;
*)

(************************************************************************)

PROCEDURE AppendAscString (VAR (*IN*) buf: ARRAY OF CARD8;  pos, N: CARDINAL;
                            VAR (*INOUT*) message: ARRAY OF CHAR);

    (* Appends N  characters from buf[pos] to message.  *)

    VAR substr: ARRAY [0..255] OF CHAR;

    BEGIN
        Copy (ADR(buf[pos]), ADR(substr), N);
        substr[N] := Nul;
        Strings.Append (substr, message);
    END AppendAscString;

(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Write N in decimal.  *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)

PROCEDURE BEval  (VAR (*IN*) buf: ARRAY OF CARD8;  pos, N: CARDINAL): CARDINAL;

    (* Treats the next N bytes, starting at buf[pos], as bigendian.  *)

    VAR result, k: CARDINAL;

    BEGIN
        result := 0;
        FOR k := 0 TO N-1 DO
            result := 256*result + buf[pos];
            INC (pos);
        END (*FOR*);
        RETURN result;
    END BEval;

(************************************************************************)
(*                             SESSION ID                               *)
(* For a full implementation, the client supplies a session ID during   *)
(* a handshake, and the server either accepts that ID or generates a    *)
(* new ID to indicate that it does not agree to resume the previous     *)
(* session.  In my present implementation I'm not cacheing old          *)
(* sessions, so I generate a unique ID every time.                      *)
(************************************************************************)

VAR NextID: sessIDtype;
    NextIDLock: Lock;

(************************************************************************)

PROCEDURE ResetNextID;

    VAR j: CARDINAL;

    BEGIN
        Obtain (NextIDLock);
        NextID[0] := maxIDsize;
        FOR j := 1 TO maxIDsize DO
            NextID[j] := 0;
        END (*FOR*);
        Release (NextIDLock);
    END ResetNextID;

(************************************************************************)

PROCEDURE IncrNextID (k: CARDINAL);

    (* Increments NextID[1..k].  We assume that the caller has          *)
    (* exclusive access to NextID.                                      *)

    BEGIN
        IF NextID[k] = MAX(CARD8) THEN
            NextID[k] := 0;
            IF k > 1 THEN
                IncrNextID (k-1);
            END (*IF*);
        ELSE
            INC (NextID[k]);
        END (*IF*);
    END IncrNextID;

(************************************************************************)

PROCEDURE GenerateSessionID (VAR (*OUT*) ID: sessIDtype);

    BEGIN
        Obtain (NextIDLock);
        ID := NextID;
        IncrNextID (maxIDsize);
        Release (NextIDLock);
    END GenerateSessionID;

(************************************************************************)
(*                        GENERATE RANDOM NUMBER                        *)
(************************************************************************)

PROCEDURE GenerateRandom (N: CARDINAL;  VAR (*OUT*) result: ARRAY OF CARD8);

    (* Puts an N-byte "random" field in result.  *)

    VAR pos: CARDINAL;

    (********************************************************************)

    PROCEDURE PutWord (A: CARDINAL);

        (* Stores the next 4 bytes of the result. *)

        VAR val: RECORD
                    CASE :BOOLEAN OF
                        FALSE:  byte: ARRAY [0..3] OF CARD8;
                        |
                        TRUE:   word: CARD32;
                    END (*CASE*);
                 END (*RECORD*);
            j: CARDINAL;

        BEGIN
            val.word := A;
            FOR j := 3 TO 0 BY -1 DO
                result[pos] := val.byte[j];
                INC (pos);
            END (*FOR*);
        END PutWord;

    (********************************************************************)

    VAR j, k: CARDINAL;

    BEGIN
        pos := 0;
        FOR j := 1 TO N DIV 4 DO
            PutWord (RandCardinal());
        END (*FOR*);
        N := N MOD 4;
        IF N > 0 THEN
            k := RandCardinal();
            FOR j := 0 TO N-1 DO
                result[pos] := k MOD 256;
                k := k DIV 256;
                INC (pos);
            END (*FOR*);
        END (*IF*);
    END GenerateRandom;

(************************************************************************)

PROCEDURE GenerateRandom32 (VAR (*OUT*) result: ARRAY OF CARD8);

    (* A variant of GenerateRandom with 32-byte result. The first four  *)
    (* bytes are Unix time, and we follow this with 28 random bytes.    *)

    VAR pos: CARDINAL;

    (********************************************************************)

    PROCEDURE PutWord (A: CARDINAL);

        (* Stores the next 4 bytes of the result. *)

        VAR val: RECORD
                    CASE :BOOLEAN OF
                        FALSE:  byte: ARRAY [0..3] OF CARD8;
                        |
                        TRUE:   word: CARD32;
                    END (*CASE*);
                 END (*RECORD*);
            j: CARDINAL;

        BEGIN
            val.word := A;
            FOR j := 3 TO 0 BY -1 DO
                result[pos] := val.byte[j];
                INC (pos);
            END (*FOR*);
        END PutWord;

    (********************************************************************)

    CONST N = 28;
    VAR j: CARDINAL;

    BEGIN
        pos := 0;
        PutWord (time());
        FOR j := 1 TO N DIV 4 DO
            PutWord (RandCardinal());
        END (*FOR*);
    END GenerateRandom32;

(************************************************************************)
(*                    SENDING DATA TO RECORD LAYER                      *)
(************************************************************************)

(*
PROCEDURE TLSsend (sess: TLSsession;  VAR (*IN*) message: ARRAY OF LOC;
                                                length: CARDINAL): CARDINAL;

    (* Sends a message of "length" bytes.  The returned value is the    *)
    (* number of bytes sent, or MAX(CARDINAL) if there was an error.    *)

    BEGIN
        (* NOT YET IMPLEMENTED *)
        RETURN MAX(CARDINAL);
    END TLSsend;
*)

(************************************************************************)
(*                         CHANGE CIPHER SPEC                           *)
(************************************************************************)

PROCEDURE ChangeCipherSpec (state: HSstate;  forinput: BOOLEAN);

    (* Commits the new encryption details that have been negotiated *)
    (* by handshaking.                                              *)

    (* Implementation note: The currently defined cipher suite which    *)
    (* requires the most material is AES_256_CBC_SHA256. It requires    *)
    (* 2 x 32 byte keys and 2 x 32 byte MAC keys, for a total 128 bytes *)
    (* of key material.  I'm allowing for 256 bytes.                    *)

    (* NOTE: this is an inefficient implementation, because it means    *)
    (* we redundantly calculate the mast_secret and key block twice.    *)
    (* (There are two ChangeCipherSpec calls, one from the client and   *)
    (* one from the server.)  We should be doing those calculations at  *)
    (* an earlier stage, probably when the client_premaster is          *)
    (* decrypted.                                                       *)

    VAR server_write: BOOLEAN;

    BEGIN
        (* Commit the keys. *)

        server_write := state^.IsServer <> forinput;
        Commit (state^.rlstate, state^.security.pending,
                        server_write, state^.security.keyblock);

    END ChangeCipherSpec;

(************************************************************************)

PROCEDURE SendChangeCipherSpec (state: HSstate);

    (* Sends a ChangeCipherSpec to the other partner, and also does *)
    (* the change at our end.                                       *)

    (* To investigate: do we need a flag to avoid a redundant double change? *)

    VAR buffer: TextType;

    BEGIN
        LogTransactionL (state^.logID, "Sending ChangeCipherSpec");
        buffer.contenttype := 20;   (* type 22 means "change cipher spec" *)
        buffer.majversion := SSLversionMajor;
        buffer.minversion := SSLversionMinor;
        buffer.length := Swap2 (1);        (* convert to BigEndian *)
        ALLOCATE (buffer.fragment, 1);
        buffer.fragment^[0] := 1;
        PutFragment (state^.rlstate, state^.logID, buffer);
        DiscardFragment (buffer);
        ChangeCipherSpec (state, FALSE);
        LogTransactionL (state^.logID, "Have sent ChangeCipherSpec");
    END SendChangeCipherSpec;

(************************************************************************)
(*                              ALERTS                                  *)
(************************************************************************)

PROCEDURE LogAlert (ID: TransactionLogID;  buffer: TextType): BOOLEAN;

    (* Returns TRUE iff not a fatal alert. *)

    VAR type, descr: CARD8;  success: BOOLEAN;
        message: ARRAY [0..127] OF CHAR;

    BEGIN
        Strings.Assign ("received ", message);
        IF buffer.fragment = NIL THEN
            Strings.Append ("buffer.fragment = NIL", message);
            success := TRUE;
        ELSE
            type := buffer.fragment^[0];
            descr := buffer.fragment^[1];
            success := type = 1;    (* warning rather than fatal *)
            IF success THEN Strings.Append ("Warning", message);
            ELSE Strings.Append ("Fatal", message);
            END (*IF*);
            IF descr = 10 THEN
                Strings.Append (" error: unexpected message", message);
            ELSE
                Strings.Append (" alert ", message);
                AppendCard (descr, message);
                CASE descr OF
                    | 40:   Strings.Append (" = handshake failure", message);
                    | 42:   Strings.Append (" = bad certificate", message);
                    | 43:   Strings.Append (" = unsupported certificate", message);
                    | 48:   Strings.Append (" = unknown CA", message);
                ELSE
                    (* Don't need explanations of other cases so far. *)
                END (*CASE*);
            END (*IF*);
        END (*IF*);
        LogTransaction (ID, message);
        RETURN success;
    END LogAlert;

(************************************************************************)
(*                            HANDSHAKES                                *)
(************************************************************************)

PROCEDURE SendHandshake (state: HSstate;  type: CARD8;
                                VAR (*INOUT*) content: ByteStringPtr;
                                  length: CARDINAL);

    (* A handshake record starts with code 22 to mean "handshake", then *)
    (* we have a two-byte version code and a two-byte overall length    *)
    (* Then we have the "fragment" field with                           *)
    (*      HandshakeType (1 byte)                                      *)
    (*      length        (3 bytes)                                     *)
    (*      further data  as supplied by the caller                     *)
    (* For efficiency the caller must provide the space for the four    *)
    (* initial bytes, which we will fill in, in the content^ record.    *)
    (* The length parameter supplied by the caller includes counting    *)
    (* these initial four bytes.                                        *)
    (* We deallocate the caller data before return.                     *)

    VAR buffer: TextType;
        j, L0: CARDINAL;

    BEGIN

        buffer.contenttype := 22;   (* type 22 means "handshake" *)
        buffer.majversion := SSLversionMajor;
        buffer.minversion := SSLversionMinor;

        (* To embed the handshake record into something suitable to     *)
        (* pass to the record layer, we have to prepend a one-byte      *)
        (* handshake type and a 3-byte length field.  At present I'm    *)
        (* doing this by requiring the caller to leave the first four   *)
        (* bytes in content^ free for our use.  There must be a better  *)
        (* way to handle a layered protocol, but for now I'll use the   *)
        (* crude obvious approach.                                      *)

        buffer.length := Swap2 (length);        (* convert to BigEndian *)
        L0 := length - 4;
        content^[0] := type;
        FOR j := 3 TO 1 BY -1 DO
            content^[j] := L0 MOD 256;
            L0 := L0 DIV 256;
        END (*FOR*);

        (*
        WriteString ("In SendHandshake, first few bytes of output are ");
        WriteHexString (content^, 0, 32, TRUE);
        WriteLn;
        *)

        buffer.fragment := content;
        IF length > 0 THEN
            IF type = 20 THEN
                (* Special case.  We have to include it it the hash if  *)
                (* this is the first handshake we have encountered,     *)
                (* either in or out, but not otherwise.                 *)
                IF CAST (ADDRESS, state^.hashctx1) <> NIL THEN
                    SHA2Update (state^.hashctx2, buffer.fragment^, length);
                END (*IF*);
            ELSIF type <> 0 THEN
                SHA2Update (state^.hashctx1, buffer.fragment^, length);
                SHA2Update (state^.hashctx2, buffer.fragment^, length);
            END (*IF*);
        END (*IF*);
        PutFragment (state^.rlstate, state^.logID, buffer);
        DiscardFragment (buffer);

    END SendHandshake;

(************************************************************************)
(*             HANDSHAKE PROTOCOL -- ClientHello EXTENSIONS             *)
(*                                                                      *)
(* Browsers that I have checked use the following extensions (hex):     *)
(*      Firefox 38.8.0    0000 0005 000A 000B 000D 0010 0023 3374 FF01  *)
(*      Firefox 59.0.2    0005 000A 000B 0010 0017 0023 0033 FF01       *)
(*                          (list possibly incomplete, there's a bug)   *)
(*      M$Edge            0005 000A 000B 000D 0010 0017 0023            *)
(*      Chrome            0005 000A 000B 000D 0010 0012 0017            *)
(*                          0023 1A1A 3A3A 7550 FF01                    *)
(*                                                                      *)
(************************************************************************)

PROCEDURE Extension0 (logID: TransactionLogID;  VAR (*IN*) record: ByteStringPtr;
                                      offset, size: CARDINAL);

    (* The server_name extension - probably not needed. *)

    VAR namelength: CARDINAL;
        nametype: CARD8;
        message: ARRAY [0..511] OF CHAR;

    BEGIN
        message := "server_name";
        WHILE size > 0 DO

            (* In each subrecord the length is redundantly given    *)
            (* twice, so we can just skip the outer length spec.    *)

            INC (offset, 2);  DEC (size, 2);
            nametype := record^[0];
            INC (offset);  DEC (size);
            namelength := BEval(record^, offset, 2);
            INC (offset, 2);  DEC (size, 2);
            Strings.Append ("    ", message);
            AppendCard (nametype, message);  Strings.Append (' ', message);
            AppendAscString (record^, offset, namelength, message);
            INC (offset, namelength);  DEC (size, namelength);
        END (*WHILE*);
        LogTransaction (logID, message);
    END Extension0;

(************************************************************************)

PROCEDURE Extension13 (logID: TransactionLogID;  VAR (*IN*) record: ByteStringPtr;
                                      offset, size: CARDINAL);

    (* The signature_algorithms extension, definitely important. *)

    VAR code, count: CARD8;
        message: ARRAY [0..511] OF CHAR;

    BEGIN
        LogTransactionL (logID, "signature_algorithms");
        (*
        WriteString ("The next few bytes in the record are: ");  WriteLn;
        WriteHexString (record^, offset, 32); WriteLn;
        *)

        size := BEval(record^, offset, 2);  INC (offset, 2);

        message := "";  count := 0;
        WHILE size > 0 DO

            Strings.Append ("  ", message);

            (* Read a pair of values, one for hash algorithm and    *)
            (* one for signature algorithm.                         *)

            code := record^[offset];
            INC (offset);  DEC (size);
            CASE code OF
                    0:  Strings.Append ("none", message);
                |   1:  Strings.Append ("md5", message);
                |   2:  Strings.Append ("sha1", message);
                |   3:  Strings.Append ("sha224", message);
                |   4:  Strings.Append ("sha256", message);
                |   5:  Strings.Append ("sha384", message);
                |   6:  Strings.Append ("sha512", message);
                ELSE    Strings.Append ("???", message);
            END (*CASE*);
            Strings.Append ('_', message);
            code := record^[offset];
            INC (offset);  DEC (size);
            CASE code OF
                    0:  Strings.Append ("anon", message);
                |   1:  Strings.Append ("rsa", message);
                |   2:  Strings.Append ("dsa", message);
                |   3:  Strings.Append ("ecdsa", message);
                ELSE    Strings.Append ("???", message);
            END (*CASE*);
            INC (count);
            IF count = 3 THEN
                LogTransaction (logID, message);
                message := "";  count := 0;
            END (*IF*);
        END (*WHILE*);
        IF count > 0 THEN
            LogTransaction (logID, message);
        END (*IF*);
    END Extension13;

(************************************************************************)

PROCEDURE AppendExtensionName (extnum: CARDINAL;  VAR (*INOUT*) message: ARRAY OF CHAR);

    VAR name: ARRAY [0..31] OF CHAR;

    BEGIN
        name[0] := Nul;
        IF extnum = 5 THEN
            name := "status_request";
        ELSIF extnum = 10 THEN
            name := "supported_groups";
        ELSIF extnum = 11 THEN
            name := "ec_point_formats";
        ELSIF extnum = 16 THEN
            name := "app_layer_prot_neg";
        ELSIF extnum = 35 THEN
            name := "session_ticket";
        ELSE
            (* Append nothing *)
        END (*IF*);
        IF name[0] <> Nul THEN
            Strings.Append (' ', message);
            Strings.Append (name, message);
        END (*IF*);
    END AppendExtensionName;

(************************************************************************)
(*                              CLIENT HELLO                            *)
(************************************************************************)

PROCEDURE ParseHandshakeExtensions (logID: TransactionLogID;
                                        VAR (*IN*) record: ByteStringPtr;
                                            offset, length: CARDINAL);

    (* Parses the extension fields in a ClientHello, but for now does   *)
    (* nothing with them.                                               *)

    VAR size, type: CARDINAL;
        message: ARRAY [0..127] OF CHAR;

    BEGIN
        length := BEval (record^, offset, 2);
        INC (offset, 2);
        WHILE length >= 4 DO

            (* Each extension begins with four bytes: an extension  *)
            (* type and a length.                                   *)

            type := BEval (record^, offset, 2);
            INC (offset, 2);
            size := BEval (record^, offset, 2);
            INC (offset, 2);
            IF type = 0 THEN
                Extension0 (logID, record, offset, size);
            ELSIF type = 13 THEN
                Extension13 (logID, record, offset, size);
            ELSE
                message := "Extension ";
                (*AppendHexString (record^, offset, 2, FALSE, message);*)
                AppendHex4 (type, message);
                Strings.Append (" length ", message);
                AppendCard (size, message);
                AppendExtensionName (type, message);
                LogTransaction (logID, message);
            END (*IF*);
            INC (offset, size);
            DEC (length, size+4);
        END (*WHILE*);
    END ParseHandshakeExtensions;

(************************************************************************)

PROCEDURE HandleClientHello (state: HSstate;  VAR (*IN*) record: ByteStringPtr;
                                      offset, length: CARDINAL): BOOLEAN;

    (* The data for this level start at record[offset].    *)

    VAR sublength, j, k, count: CARDINAL;
        sessionID: ARRAY [0..31] OF CARD8;
        name: ARRAY [0..31] OF CHAR;
        code1, code2: CARD8;
        suite, newsuite: CipherSuite;
        method, newmethod: CompressionMethod;
        message: ARRAY [0..127] OF CHAR;

    BEGIN
        (*
        WriteString ("In HandleClientHello, first few bytes of input are ");
        WriteHexString (record^, 0, 32, TRUE);
        WriteLn;
        *)
        (*
        message := "In HandleClientHello, initially length = ";
        AppendCard (length, message);
        LogTransaction (state^.logID, message);
        *)

        (* VERSION - which we are currently not checking for compatibility *)

        (* The first two bytes should be the version.  Note that I  *)
        (* am updating length to show how many bytes are left in    *)
        (* this Hello message.                                      *)

        (*
        WriteString ("Protocol version ");  WriteCard (record^[offset]);
        WriteChar ('.');  WriteCard (record^[offset+1]);  WriteLn;
        *)
        DEC (length, 2);

        (* CLIENT RANDOM SEED *)

        (* Copy client random seed. *)

        Copy (ADR(record^[offset+2]), ADR(state^.security.client_random), 32);
        (*
        Strings.Assign ("client_random = ", message);
        AppendHexString (state^.security.client_random, 0, 32, FALSE, message);
        LogTransaction (state^.logID, message);
        *)
        DEC (length, 32);

        (* First four bytes of seed should be Unix time. *)

        (*
        unixtime := 0;
        FOR j := 0 TO 3 DO
            unixtime := 256*unixtime + state^.security.client_random[j];
        END (*FOR*);
        WriteString ("Unix time = ");  WriteCard (unixtime);  WriteLn;
        *)

        (* SESSION ID - ignored in this version *)

        (* Client's session ID.  The first byte is the length.  *)

        j := offset+34;
        sublength := record^[j];  INC (j);
        DEC (length, sublength+1);
        k := 0;
        WHILE sublength > 0 DO
            sessionID[k] := record^[j];  INC(j);  INC(k);
            DEC (sublength);
        END (*WHILE*);
        IF k < 32 THEN
            sessionID[k] := 0;
        END (*IF*);
        Strings.Assign ("Session ID = ", message);
        IF k = 0 THEN
            Strings.Append ("''", message);
        ELSE
            AppendHexString (sessionID, 0, k, FALSE, message);
        END (*IF*);
        LogTransaction (state^.logID, message);

        (* CIPHER SUITES *)

        (* Now we have to get a list of the cipher suites acceptable to the client. *)
        (* Each cipher suite code is two bytes long. *)

        suite := NullSuite;
        sublength := 256*record^[j] + record^[j+1];
        INC (j, 2);
        DEC (length, sublength + 2);
        sublength := sublength DIV 2;
        IF ListCipherSuites THEN
            LogTransactionL (state^.logID, "Client will accept cipher suites:");
        END (*IF*);
        message := "";  count := 0;
        FOR k := 0 TO sublength-1 DO
            code1 := record^[j];  INC (j);
            code2 := record^[j];  INC (j);
            IF ListCipherSuites THEN
                Strings.Append ("    ", message);
                AppendHex (code1, message);
                Strings.Append (' ', message);
                AppendHex (code2, message);
            END (*IF*);
            IF ListCipherSuites OR (suite = NullSuite) THEN
                newsuite := AcceptCipherSuite(code1, code2);
                IF (suite = NullSuite) AND (newsuite <> NullSuite) THEN
                    suite := newsuite;
                END (*IF*);
                IF ListCipherSuites THEN
                    IF newsuite <> NullSuite THEN
                        IF suite = newsuite THEN
                            Strings.Append ("   accepted", message);
                        ELSE
                            Strings.Append ("   acceptable", message);
                        END (*IF*);
                    ELSE
                        Strings.Append ("   rejected", message);
                    END (*IF*);
                    INC (count);
                    IF count = 2 THEN
                        LogTransaction (state^.logID, message);
                        message := "";  count := 0;
                    END (*IF*);
                END (*IF*);
            END (*IF*);
        END (*FOR*);
        IF ListCipherSuites AND (count > 0) THEN
            LogTransaction (state^.logID, message);
        END (*IF*);
        state^.security.pending := suite;
        Strings.Assign ("ciphersuite = ", message);
        NameOfSuite (suite, name);
        Strings.Append (name, message);
        LogTransaction (state^.logID, message);

        (*
        message := "Just before getting compression algorithms, j = ";
        AppendCard (j, message);
        Strings.Append (" and length = ", message);
        AppendCard (length, message);
        LogTransaction (state^.logID, message);
        *)

        (* COMPRESSION ALGORITHMS *)

        (* And then the compression methods. *)

        sublength := record^[j];
        INC (j);
        DEC (length, sublength+1);

        method := nocompression;
        FOR k := 1 TO sublength DO
            code1 := record^[j];  INC (j);
            newmethod := CompressionSupported(code1);
            IF (method = nocompression) AND (newmethod <> nocompression) THEN
                method := newmethod;
            END (*IF*);
        END (*FOR*);
        state^.security.compression_algorithm := method;

        (* EXTENSIONS *)

        IF length > 0 THEN

            ParseHandshakeExtensions (state^.logID, record, j, length);

        END (*IF*);

        RETURN TRUE;

    END HandleClientHello;

(************************************************************************)

PROCEDURE SendClientHello (state: HSstate): BOOLEAN;

    (* Send the client hello to the server.         *)
    (* The data for this level start at record[0].  *)

    CONST
        InitialSize = 128;
        client_hello_code = 1;

    VAR recsize, length: CARDINAL;
        name: ARRAY [0..31] OF CHAR;
        record: ByteStringPtr;
        suite: CipherSuite;
        message: ARRAY [0..127] OF CHAR;
        code1, code2: CARD8;

    BEGIN
        LogTransactionL (state^.logID, "Sending client hello");
        recsize := InitialSize;
        ALLOCATE (record, recsize);
        length := 4;            (* Initial space reserved for SendHandshake *)

        (* VERSION *)

        (* The first two bytes should be the version.  *)

        Checksize (record, recsize, length, 2);
        record^[length] := SSLversionMajor;  INC (length);
        record^[length] := SSLversionMinor;  INC (length);

        (* CLIENT RANDOM SEED *)

        (* Create and send client random seed. *)

        GenerateRandom32 (state^.security.client_random);
        Checksize (record, recsize, length, 32);
        Copy (ADR(state^.security.client_random), ADR(record^[length]), 32);
        (*
        Strings.Assign ("client_random = ", message);
        AppendHexString (state^.security.client_random, 0, 32, FALSE, message);
        LogTransaction (state^.logID, message);
        *)
        INC (length, 32);

        (* SESSION ID - just set to null string in this version *)

        (* Client's session ID.  The first byte is the length.  *)

        Checksize (record, recsize, length, 1);
        record^[length] := 0;  INC (length);

        (* CIPHER SUITES *)

        (* Now we have to get a list of the cipher suites acceptable to the client. *)
        (* Each cipher suite code is two bytes long. *)
        (* In this version I'll just list one suite, for testing.    *)

        (*suite := NullSuite;*)
        (*suite := TLS_RSA_WITH_AES_256_CBC_SHA256;*)
        suite := TLS_RSA_WITH_3DES_EDE_CBC_SHA;       (* google.com accepts this one *)
        CipherSuiteToCode (suite, code1, code2);
        Checksize (record, recsize, length, 4);
        record^[length] := 0;  INC (length);        (* length of list (2 bytes) *)
        record^[length] := 2;  INC (length);
        record^[length] := code1;  INC (length);
        record^[length] := code2;  INC (length);
        Strings.Assign ("ciphersuite = ", message);
        NameOfSuite (suite, name);
        Strings.Append (name, message);
        LogTransaction (state^.logID, message);

        (* COMPRESSION ALGORITHMS *)

        (* And then the compression methods. *)
        (* For now we support only "no compression". *)

        state^.security.compression_algorithm := nocompression;
        Checksize (record, recsize, length, 2);
        record^[length] := 1;  INC (length);        (* length of list (1 byte) *)
        record^[length] := ORD(state^.security.compression_algorithm);  INC (length);

        (* EXTENSIONS *)

        (* No extensions supported in this version.  *)

        Resize (record, recsize, length);

        message := "Sending client hello, ";
        AppendCard (length, message);
        Strings.Append (" bytes.", message);
        LogTransaction (state^.logID, message);
        (*
        WriteString ("First few bytes of output are ");
        WriteHexString (record^, 0, 51, TRUE);
        WriteLn;
        *)

        SendHandshake (state, client_hello_code, record, length);

        RETURN TRUE;

    END SendClientHello;

(************************************************************************)
(*                            KEY EXCHANGE                              *)
(************************************************************************)

PROCEDURE GenerateKeyBlock (state: HSstate);

    (* Calculates the master secret and the block of key information    *)
    (* required for a pending change of cipher suite.  This can be      *)
    (* done once ClientKeyExchange has supplied a non-encrypted         *)
    (* client_premaster.                                                *)

    VAR seed: ARRAY [0..63] OF CARD8;
        blocksize: CARDINAL;

    BEGIN
        (* Note that the seed for the following two calculations is     *)
        (* the same except that the order is swapped.  I can't yet      *)
        (* guess whether this is an error in the standard.              *)

        Copy (ADR(state^.security.client_random), ADR(seed), 32);
        Copy (ADR(state^.security.server_random), ADR(seed[32]), 32);
        PRF (state^.security.prf_algorithm,
                48, state^.security.client_premaster,
                "master secret", 64, seed, 48, state^.security.master_secret);

        (* Generate the key block. *)

        blocksize := KeyBlockSize (state^.security.pending);
        Copy (ADR(state^.security.server_random), ADR(seed), 32);
        Copy (ADR(state^.security.client_random), ADR(seed[32]), 32);
        PRF (state^.security.prf_algorithm,
                48, state^.security.master_secret,
                "key expansion", 64, seed,
                 blocksize, state^.security.keyblock);

    END GenerateKeyBlock;

(************************************************************************)

PROCEDURE HandleClientKeyExchange (state: HSstate;  VAR (*IN*) record: ByteStringPtr;
                                      offset, length: CARDINAL): BOOLEAN;

    (* The data for this level start at record[offset].                 *)
    (* In the present version the only supported key exchange algorithm *)
    (* is RSA, so what the client sends is the encrypted premaster      *)
    (* secret.  This is a two-byte version, followed by a 46-byte       *)
    (* random value.  The version is part of the premaster secret.      *)

    (* I am temporarily hardcoding the key file name. Will later have   *)
    (* to look this up in INI file.                                     *)

    (*CONST keyfile = "C:\etc\pki\tls\private\testcert.priv.key";*)

    VAR success: BOOLEAN;
        message: ARRAY [0..2047] OF CHAR;
        decrypted: ARRAY [0..2047] OF CARD8;
        key: RSAKeys.RSAKeyType;
        encrypted: ByteStr;
        S, T: BN;

    BEGIN
        success := TRUE;

        (* CLIENT PREMASTER SECRET *)

        (* Copy (encrypted) client premaster seed. *)
        (* The record starts with two length bytes.  *)

        length := 256*record^[offset] + record^[offset+1];
        MakeBS (encrypted, length);
        Copy (ADR(record^[offset+2]), encrypted.data, length);

        (*
        message := "encrypted client_premaster ";
        AppendHexString (encrypted.data^, 0, length, TRUE, message);
        LogTransaction (state^.logID, message);
        *)

        (* Decrypt the seed. *)

        key := TLSRSA.LoadRSAPrivateKey (PrivateKeyFile);
        S := BigNum.BinToBN (encrypted.data^, 0, length);
        T := RSA.PrivKeyEncode (S, key);
        length := BigNum.BNtoBytes (T, decrypted (*state^.security.client_premaster*));

        (*
        message := "decrypted block ";
        AppendHexString (decrypted, 0, length, TRUE, message);
        LogTransaction (state^.logID, message);
        *)

        (* Extract the last 48 bytes from this block.  The rest is  *)
        (* padding, which we can ignore.                            *)

        Copy (ADR(decrypted[length-48]), ADR(state^.security.client_premaster), 48);

        message := "decrypted client_premaster ";
        AppendHexString (state^.security.client_premaster, 0, 48, TRUE, message);
        LogTransaction (state^.logID, message);

        (* VERSION *)

        (* The first two bytes should be the version.  At present I'm   *)
        (* not bothering to check this.                                 *)

        (*
        WriteString ("Premaster protocol version ");  WriteCard (record^[offset]);
        WriteChar ('.');  WriteCard (record^[offset+1]);  WriteLn;
        *)

        (* We now have enough information to calculate the master   *)
        (* secret and the block of key material.                    *)

        GenerateKeyBlock (state);

        RETURN success;

    END HandleClientKeyExchange;

(************************************************************************)

PROCEDURE SendClientKey (state: HSstate): BOOLEAN;

    (* Send the client premaster information to the server.             *)
    (* In the present version the only supported key exchange algorithm *)
    (* is RSA, so what the client sends is the encrypted premaster      *)
    (* secret.  This is a two-byte version, followed by a 46-byte       *)
    (* random value which is the premaster secret.                      *)

    (* TODO: add some padding to the record. *)
    (* TODO: encrypt premaster before sending it *)

    CONST
        InitialSize = 128;
        client_keyexchange_code = 16;

    VAR recsize, length: CARDINAL;
        name: ARRAY [0..31] OF CHAR;
        record: ByteStringPtr;
        suite: CipherSuite;
        message: ARRAY [0..127] OF CHAR;
        code1, code2: CARD8;

    BEGIN
        LogTransactionL (state^.logID, "Sending client hello");
        recsize := InitialSize;
        ALLOCATE (record, recsize);
        length := 4;            (* Initial space reserved for SendHandshake *)

        (* VERSION *)

        (* The first two bytes should be the version.  *)

        Checksize (record, recsize, length, 2);
        record^[length] := SSLversionMajor;  INC (length);
        record^[length] := SSLversionMinor;  INC (length);

        (* CLIENT PREMASTER SECRET *)

        (* For RSA, this is a two-byte version followed by 46 random bytes. *)

        GenerateRandom (48, state^.security.client_premaster);
        Checksize (record, recsize, length, 48);
        Copy (ADR(state^.security.client_premaster), ADR(record^[length]), 48);
        INC (length, 48);

        (* We now have enough information to calculate the master   *)
        (* secret and the block of key material.                    *)

        GenerateKeyBlock (state);

        (******* INCOMPLETE CODE ********)
        (* We must encrypt the client premaster before sending it.  *)

        SendHandshake (state, client_keyexchange_code, record, length);

        RETURN TRUE;

    END SendClientKey;

(************************************************************************)
(*                             SERVER HELLO                             *)
(************************************************************************)

PROCEDURE CopyBytes (VAR (*IN*) src: ARRAY OF LOC;  N: CARDINAL;
                      record: ByteStringPtr;  VAR (*INOUT*) pos: CARDINAL);

    (* Copies N bytes from src to record^[pos], updates pos. *)

    VAR j: CARDINAL;

    BEGIN
        FOR j := 0 TO N-1 DO
            record^[pos] := CAST (CARD8,src[j]);
            INC (pos);
        END (*FOR*);
    END CopyBytes;

(************************************************************************)

PROCEDURE AssembleServerHello (state: HSstate;
                        VAR (*OUT*) record: ByteStringPtr): CARDINAL;

    (* Puts together the content of a server hello, returns its length. *)
    (* This content consists of                                         *)
    (*      overhead        (4 bytes)                                   *)
    (*      version         (2 bytes)                                   *)
    (*      random seed     (32 bytes)                                  *)
    (*      session ID      (1-byte length, then up to 32 bytes of ID)  *)
    (*      ciphersuite     (2 bytes)                                   *)
    (*      compression method  (1 byte)                                *)
    (* That adds up to 42 plus the session ID length.                   *)
    (* Extensions can follow, but for now I'll leave those out.         *)

    VAR pos, amount: CARDINAL;
        code1, code2: CARD8;

    BEGIN
        amount := 42 + state^.sessID[0];
        ALLOCATE (record, amount);
        record^[4] := SSLversionMajor;  record^[5] := SSLversionMinor;
        pos := 6;
        CopyBytes (state^.security.server_random, 32, record, pos);
        CopyBytes (state^.sessID, state^.sessID[0]+1, record, pos);
        CipherSuiteToCode (state^.security.pending, code1, code2);
        record^[pos] := code1;  INC(pos);
        record^[pos] := code2;  INC(pos);
        record^[pos] := ORD(nocompression);  INC(pos);
        IF pos <> amount THEN
            WriteString ("Allocation error in AssembleServerHello.");
            WriteLn;
        END (*IF*);

        (* Only signature algorithm I can support at present is RSA     *)
        (* with SHA256, because that's all the certificate has.  But in *)
        (* any case the standard says that a server MUST NOT send the   *)
        (* signature_algorithms extension, although it must be able to  *)
        (* receive it.                                                  *)

        RETURN amount;

    END AssembleServerHello;

(************************************************************************)

PROCEDURE SendServerHello (state: HSstate);

    (* A handshake record starts with code 22 to mean "handshake", then *)
    (* we have a two-byte version code and a two-byte overall length    *)
    (* Then we have the "fragment" field with                           *)
    (*      HandshakeType (1 byte, value 2 for server hello)            *)
    (*      length        (3 bytes)                                     *)
    (*      content       (depends on handshake type)                   *)

    CONST server_hello_code = 2;

    VAR p: ByteStringPtr;
        length: CARDINAL;
        message: ARRAY [0..127] OF CHAR;

    BEGIN

        GenerateSessionID (state^.sessID);
        GenerateRandom32 (state^.security.server_random);
        (*
        Strings.Assign ("server_random = ", message);
        AppendHexString (state^.security.server_random, 0, 32, FALSE, message);
        LogTransaction (state^.logID, message);
        *)
        length := AssembleServerHello (state, p);
        SendHandshake (state, server_hello_code, p, length);

    END SendServerHello;

(************************************************************************)

PROCEDURE ReceiveServerHello (state: HSstate;  record: ByteStringPtr;
                                    offset, length: CARDINAL): BOOLEAN;

    (* Receive the server_hello handshake.                              *)

    (*      overhead        (4 bytes, handled by caller)                *)
    (*      version         (2 bytes)                                   *)
    (*      random seed     (32 bytes)                                  *)
    (*      session ID      (1-byte length, then up to 32 bytes of ID)  *)
    (*      ciphersuite     (2 bytes)                                   *)
    (*      compression method  (1 byte)                                *)

    VAR message: ARRAY [0..127] OF CHAR;
        name: ARRAY [0..31] OF CHAR;
        j, k, sublength: CARDINAL;
        code1, code2: CARD8;
        sessionID: ARRAY [0..31] OF CARD8;

    BEGIN
        LogTransactionL (state^.logID, "Receiving server hello");

        (* VERSION *)

        (* The first two bytes should be the version.  Note that I  *)
        (* am updating length to show how many bytes are left in    *)
        (* this Hello message.                                      *)

        (*
        WriteString ("Protocol version ");  WriteCard (record^[offset]);
        WriteChar ('.');  WriteCard (record^[offset+1]);  WriteLn;
        *)
        DEC (length, 2);

        (* SERVER RANDOM SEED *)

        (* Accept server random seed. *)

        Copy (ADR(record^[offset+2]), ADR(state^.security.server_random), 32);
        (*
        Strings.Assign ("server_random = ", message);
        AppendHexString (state^.security.server_random, 0, 32, FALSE, message);
        LogTransaction (state^.logID, message);
        *)
        DEC (length, 32);

        (* Client's session ID.  The first byte is the length.  *)

        j := offset+34;
        sublength := record^[j];  INC (j);
        DEC (length, sublength+1);
        k := 0;
        WHILE sublength > 0 DO
            sessionID[k] := record^[j];  INC(j);  INC(k);
            DEC (sublength);
        END (*WHILE*);
        IF k < 32 THEN
            sessionID[k] := 0;
        END (*IF*);
        Strings.Assign ("Session ID = ", message);
        IF k = 0 THEN
            Strings.Append ("''", message);
        ELSE
            AppendHexString (sessionID, 0, k, FALSE, message);
        END (*IF*);
        LogTransaction (state^.logID, message);
        offset := j;

        (* CIPHER SUITE *)

        code1 := record^[offset];  INC (offset);  DEC(length);
        code2 := record^[offset];  INC (offset);  DEC(length);
        Strings.Assign ("ciphersuite = ", message);
        state^.security.pending := CodeToCipherSuite (code1, code2);
        NameOfSuite (state^.security.pending, name);
        Strings.Append (name, message);
        LogTransaction (state^.logID, message);

        (* COMPRESSION *)

        (* The compression method code is a single byte. *)

        code1 := record^[offset];  INC (offset);  DEC (length);
        state^.security.compression_algorithm := CompressionSupported (code1);

        IF length = 0 THEN
            LogTransactionL (state^.logID, "Finished processing server hello");
            RETURN TRUE;
        ELSE
            LogTransactionL (state^.logID, "Unexpected record length in ReceiveServerHello");
            RETURN FALSE;
        END (*IF*);

    END ReceiveServerHello;

(************************************************************************)
(*                        SERVER HELLO DONE                             *)
(************************************************************************)

PROCEDURE ReceiveServerHelloDone (state: HSstate): BOOLEAN;

    (* Receive the server_hello_done handshake. *)

    BEGIN
        (* There is nothing to do here apart from checking that the     *)
        (* server_hello_done message has arrived, and the caller has    *)
        (* seen that; otherwise we wouldn't have been called.           *)

        RETURN TRUE;

    END ReceiveServerHelloDone;

(************************************************************************)

PROCEDURE SendServerHelloDone (state: HSstate);

    (* A handshake record starts with code 22 to mean "handshake", then *)
    (* we have a two-byte version code and a two-byte overall length    *)
    (* Then we have the "fragment" field with                           *)
    (*      HandshakeType (1 byte, value 14 for server hello done)      *)
    (*      length        (3 bytes)                                     *)
    (*      content       (empty, in the hello_done case)               *)

    CONST server_hello_done_code = 14;

    VAR p: ByteStringPtr;

    BEGIN

        ALLOCATE (p, 4);
        RecordTraceOn;
        SendHandshake (state, server_hello_done_code, p, 4);
        RecordTraceOff;

    END SendServerHelloDone;

(************************************************************************)
(*                         CERTIFICATE EXCHANGE                         *)
(************************************************************************)

PROCEDURE SendServerCertificate (state: HSstate);

    (* Sends our certificate to the client.  In this version I use a    *)
    (* test certificate with a fixed filename, and there is only one    *)
    (* certificate in the chain.                                        *)

    (* The required format is a 3-byte total length; then, for each     *)
    (* certificate, a 3-byte length followed by the certificate itself. *)
    (* That suggests that I need to define a new data type that allows  *)
    (* linking of nested structures, but for now I can't think of an    *)
    (* elegant way to do this.                                          *)

    VAR pdata: ByteStringPtr;
        amount: CARDINAL;

    CONST certificate_code = 11;

    BEGIN
        AssembleOurCertificates (amount, pdata, state^.logID);
        RecordTraceOn;
        SendHandshake (state, certificate_code, pdata, amount);
        RecordTraceOff;
    END SendServerCertificate;

(************************************************************************)
(*                              FINISHED                                *)
(* This is sent just before exiting from handshake mode and moving on   *)
(* to sending and receiving application data. Usually the server sends  *)
(* it first, just after issuing a ChangeCipherSpec. In that case the    *)
(* client should respond by sending its own ChangeCipherSpec and then   *)
(* its own Finished message.  There are some situations where the       *)
(* client is the one to send the first Finished.                        *)
(************************************************************************)

PROCEDURE CalculateFinishedData (state: HSstate;  fromserver: BOOLEAN;
                                  resultlength: CARDINAL;
                                  VAR (*OUT*) result: ARRAY OF CARD8);

    (* result should match the value in the Finished message. *)

    VAR label: ARRAY [0..15] OF CHAR;
        hash: ARRAY [0..311] OF CARD8;

    BEGIN
        (* Make sure to use the two hash results in *)
        (* the correct order.                       *)

        IF CAST(ADDRESS, state^.hashctx1) <> NIL THEN
            SHA2Final (state^.hashctx1, hash);
        ELSE
            SHA2Final (state^.hashctx2, hash);
        END (*IF*);
        IF fromserver THEN
            label := "server finished";
        ELSE
            label := "client finished";
        END (*IF*);
        (* The accumulated hash is the seed *)
        PRF (state^.security.prf_algorithm,
                48, state^.security.master_secret,
                label, 32, hash, resultlength, result);
    END CalculateFinishedData;

(************************************************************************)

PROCEDURE ReceiveFinished (state: HSstate;  buffer: TextType): BOOLEAN;

    (* See page 63. The data transferred with the message, 12 bytes     *)
    (* long for all cipher suites I handle so far, was created          *)
    (* with a PRF call, which uses a hash of all handshake messages     *)
    (* (in both directions, I presume) up until now, but not including  *)
    (* HelloRequest messages.  That means that I have to add some hash  *)
    (* calculations at the appropriate point.  As I understand it, the  *)
    (* calculation is the same for the sender and receiver of the       *)
    (* Finished message.                                                *)

    VAR length, pos: CARDINAL;
        message: ARRAY [0..79] OF CHAR;
        digest: ARRAY [0..31] OF CARD8;

    BEGIN
        message := "Finished message received, ";
        pos := 1;
        length := Get3Bytes (buffer.fragment^, pos);
        AppendCard (length, message);
        Strings.Append (" bytes", message);
        LogTransaction (state^.logID, message);
        message := "received ";
        AppendHexString (buffer.fragment^, 4, length, TRUE, message);
        LogTransaction (state^.logID, message);

        (* Work out what those bytes should have been. *)

        CalculateFinishedData (state, TRUE, length, digest);
        message := "expected ";
        AppendHexString (digest, 0, length, TRUE, message);
        LogTransaction (state^.logID, message);

        RETURN TRUE;
    END ReceiveFinished;

(************************************************************************)

PROCEDURE SendFinished (state: HSstate);

    (* Sends out a "finished" message. *)

    CONST length = 12;      (* True for all the suites we support *)
        finishedcode = 20;

    VAR p: ByteStringPtr;

    BEGIN
        LogTransactionL (state^.logID, "Sending Finished handshake message");
        ALLOCATE (p, length);
        CalculateFinishedData (state, FALSE, length, p^);
        SendHandshake (state, finishedcode, p, length);
        LogTransactionL (state^.logID, "Have sent Finished message");
    END SendFinished;

(************************************************************************)
(*                 RECEIVE HANDSHAKE MESSAGE FROM PEER                  *)
(************************************************************************)

PROCEDURE ReceivePartnerHandshake (state: HSstate;
                                    VAR (*OUT*) done: BOOLEAN): BOOLEAN;

    VAR type: CARD8;  success: BOOLEAN;
        j, length, L: CARDINAL;
        buffer: TextType;
        message: ARRAY [0..127] OF CHAR;
        modulus, exponent: BN;

    (* A handshake record starts with code 22 to mean "handshake", then *)
    (* we have a two-byte version code and a two-byte length.  Those    *)
    (* initial five bytes are loaded into the appropriate fields of     *)
    (* buffer by procedure GetFragment.  After this we have the         *)
    (* "fragment" field with                                            *)
    (*      HandshakeType (1 byte)                                      *)
    (*      length        (3 bytes)                                     *)
    (*      content       (depends on handshake type)                   *)

    (* The expected sequence, as seen by the client end, is:            *)
    (*      send client Hello                                           *)
    (*      receive server Hello                                        *)
    (*      (maybe) receive server certificate                          *)
    (*      (maybe) receive server key exchange                         *)
    (*      (maybe) receive certificate request from server             *)
    (*      receive server HelloDone                                    *)
    (*      (maybe) send client certificate                             *)
    (*      send client KeyExchange                                     *)
    (*      (maybe) send CertificateVerify                              *)
    (*      send ChangeCipherSpec (not a handshake message)             *)
    (*      send Finished                                               *)
    (*      receive ChangeCipherSpec                                    *)
    (*      receive Finished                                            *)
    (* After this, application data can flow.                           *)

    BEGIN
        done := FALSE;
        (*RecordTraceOn (state^.logID);*)
        success := GetFragment (state^.rlstate, state^.logID, buffer);
        RecordTraceOff;
        length := Swap2 (buffer.length);

        IF buffer.contenttype = 20 THEN

            (* Change cipher spec. *)

            ChangeCipherSpec (state, TRUE);
            success := TRUE;

        ELSIF buffer.contenttype = 21 THEN

            (* Alert *)

            success := LogAlert (state^.logID, buffer);

        ELSIF buffer.contenttype = 22 THEN

            (* Handshake *)

            IF buffer.fragment = NIL THEN
                type := 0;
                length := 0;
                L := 0;
            ELSE
                type := buffer.fragment^[0];

                (* Interpret the three-byte length. *)

                j := 1;
                length := Get3Bytes (buffer.fragment^, j);
                L := length + 4;
            END (*IF*);

            (* Update handshake hash.  We update both instances except in   *)
            (* the special case of a Finished message.                      *)

            IF L > 0 THEN
                IF type = 20 THEN
                    (* Include this in second hash iff it is the first  *)
                    (* Finished record.                                 *)
                    IF CAST (ADDRESS, state^.hashctx1) <> NIL THEN
                        SHA2Update (state^.hashctx2, buffer.fragment^, L);
                    END (*IF*);
                ELSE
                    IF type <> 0 THEN
                        SHA2Update (state^.hashctx1, buffer.fragment^, L);
                        SHA2Update (state^.hashctx2, buffer.fragment^, L);
                    END (*IF*);
                END (*IF*);
            END (*IF*);

            CASE type OF

                | 0:    (* Hello request, only ever sent from server to client. *)

                        NYI (state^.logID, "Hello request");

                | 1:    (* Client Hello received, so this is server code. *)

                        (*
                        message := "Just before calling HandleClientHello, length = ";
                        AppendCard (length, message);
                        LogTransaction (state^.logID, message);
                        *)
                        success := HandleClientHello (state, buffer.fragment, 4, length);
                        IF success THEN
                            LogTransactionL (state^.logID, "Sending server hello");
                            SendServerHello (state);
                            LogTransactionL (state^.logID, "Sending server certificate");
                            SendServerCertificate (state);
                            LogTransactionL (state^.logID, "Sending hello done");
                            SendServerHelloDone (state);

                            (* At this stage we expect the client   *)
                            (* to send more handshake records:      *)
                            (*  certificate (optional)              *)
                            (*  client key exchange                 *)
                            (*  certificate verify (optional)       *)
                            (*  change cipher spec                  *)
                            (* (That last one is not a handshake    *)
                            (* message, but a different content     *)
                            (* type.)                               *)
                            (* Then we expect a "finished"          *)
                            (* handshake message.                   *)

                            REPEAT
                                success := ReceivePartnerHandshake (state, done);
                            UNTIL done OR NOT success;
                        ELSE done := TRUE;
                        END (*IF*);

                |  2:   (* Server hello *)

                        success := ReceiveServerHello (state, buffer.fragment, 4, length);

                (*  3..10 do not appear in RFC 5246 *)

                | 11:   (* Certificate *)

                        success := AcceptCertificates (state^.logID,
                                      buffer.fragment, 4, length, modulus, exponent);

                (* 12 = server key exchange, not implemented here. *)
                (* 13 = certificate request, not implemented here. *)

                | 14:   (* Server Hello done *)

                        LogTransactionL (state^.logID, "server_hello_done received");
                        success := ReceiveServerHelloDone (state);
                        done := TRUE;

                (* 15 = certificate verify, not implemented here. *)

                | 16:
                        success := HandleClientKeyExchange(state, buffer.fragment, 4, length);

                | 20:   (* Finished. *)

                        success := ReceiveFinished (state, buffer);

                        (* Are we there yet? *)

                        IF CAST(ADDRESS, state^.hashctx2) <> NIL THEN
                            SendChangeCipherSpec (state);
                            SendFinished (state);
                        END (*IF*);
                        done := TRUE;

                |
                ELSE
                        WriteString ("Unhandled handshake type");  WriteLn;
                        WriteCard (type);  WriteLn;
                        success := FALSE;
            END (*CASE*);
        ELSE
            Strings.Assign ("Received record, type ", message);
            AppendCard (buffer.contenttype, message);
            Strings.Append (", is not a handshake record", message);
            LogTransaction (state^.logID, message);
            success := FALSE;
        END (*IF*);

        IF NOT success THEN
            DiscardFragment (buffer);
        END (*IF*);
        RETURN success;

    END ReceivePartnerHandshake;

(************************************************************************)

PROCEDURE CreateStateRecord (S: Socket;  as_server: BOOLEAN;
                                    rlstate: RLState;
                                    logID: TransactionLogID): HSstate;

    (* Sets up the initial handshake protocol state.    *)

    VAR result: HSstate;

    BEGIN
        NEW (result);
        result^.IsServer := as_server;
        result^.sock := S;
        result^.rlstate := rlstate;
        result^.logID := logID;
        WITH result^.security DO
            prf_algorithm           := tls_prf_sha256;
            bulk_cipher_algorithm   := nocipher;
            cipher_type             := stream;
            enc_key_length          := 0;
            block_length            := 0;
            fixed_iv_length         := 0;
            record_iv_length        := 0;
            mac_algorithm           := null;
            mac_length              := 0;
            mac_key_length          := 0;
            compression_algorithm   := nocompression;
            master_secret           := Zero48;
            client_random           := Zero32;
            server_random           := Zero32;
            pending                 := NullSuite;
            ciphersuite             := NullSuite;
        END (*WITH*);
        RETURN result;
    END CreateStateRecord;

(************************************************************************)

PROCEDURE DoHandshake (S: Socket;  as_server: BOOLEAN;
                                    rlstate: RLState;
                                    logID: TransactionLogID): HSstate;

    (* A handshake record starts with code 22 (16H) to mean "handshake",*)
    (* then we have a two-byte version code and a two-byte length.      *)
    (* Those initial five bytes are loaded into the appropriate fields  *)
    (* of buffer by procedure GetFragment.  After this we have the      *)
    (* "fragment" field with                                            *)
    (*      HandshakeType (1 byte)                                      *)
    (*      length        (3 bytes)                                     *)
    (*      content       (depends on handshake type)                   *)

    (* The expected sequence, as seen by the server end, is:            *)
    (*      receive client Hello                                        *)
    (*      send server Hello                                           *)
    (*      (maybe) send server certificate                             *)
    (*      (maybe) send server key exchange                            *)
    (*      (maybe) send certificate request from server                *)
    (*      send server HelloDone                                       *)
    (*      (maybe) receive client certificate                          *)
    (*      receive client KeyExchange                                  *)
    (*      (maybe) receive CertificateVerify                           *)
    (*      receive ChangeCipherSpec (not a handshake message)          *)
    (*      receive Finished                                            *)
    (*      send ChangeCipherSpec                                       *)
    (*      send Finished                                               *)
    (* After this, application data can flow.                           *)

    (* From the viewpoint of the client, this is:                       *)
    (*      send client Hello                                           *)
    (*      receive server Hello                                        *)
    (*      (maybe) receive server certificate                          *)
    (*      (maybe) receive server key exchange                         *)
    (*      (maybe) receive certificate request from server             *)
    (*      receive server HelloDone                                    *)
    (*      (maybe) send client certificate                             *)
    (*      send client KeyExchange                                     *)
    (*      (maybe) send CertificateVerify                              *)
    (*      send ChangeCipherSpec (not a handshake message)             *)
    (*      send Finished                                               *)
    (*      receive ChangeCipherSpec                                    *)
    (*      receive Finished                                            *)
    (* After this, application data can flow.                           *)

    VAR done, success: BOOLEAN;
        state: HSstate;

    BEGIN
        state := CreateStateRecord (S, as_server, rlstate, logID);
        state^.hashctx1 := SHA256Init();
        state^.hashctx2 := SHA256Init();

        done := FALSE;
        IF as_server THEN
            success := TRUE;
        ELSE
            success := SendClientHello (state);
        END (*IF*);
        WHILE success AND NOT done DO
            success := ReceivePartnerHandshake (state, done);
        END (*WHILE*);
        IF NOT success THEN
            DISPOSE (state);
        END (*IF*);
        RETURN state;
    END DoHandshake;

(************************************************************************)
(*                            INITIALISATION                            *)
(************************************************************************)

PROCEDURE LoadTLSINIData (VAR (*IN*) ININame: ARRAY OF CHAR);

    (* Loads some INI file parameters. *)

    VAR hini: HINI;
        SYSapp: ARRAY [0..7] OF CHAR;

    BEGIN
        PrivateKeyFile := "";
        SYSapp := '$SYS';
        hini := OpenINIFile(ININame);
        IF INIValid (hini) THEN
            EVAL (INIGetString (hini, SYSapp, 'PrivateKey', PrivateKeyFile));
            CloseINIFile (hini);
        END (*IF*);
        GetCertificateNames (ININame);
    END LoadTLSINIData;

(************************************************************************)

BEGIN
    CreateLock (NextIDLock);
    ResetNextID;
END TLSHandshake.

