IMPLEMENTATION MODULE TLSRSA;

        (********************************************************)
        (*                                                      *)
        (*                 TLS-RSA private key                  *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            14 October 2024                 *)
        (*  Last edited:        15 October 2024                 *)
        (*  Status:             Just started                    *)
        (*                                                      *)
        (********************************************************)


IMPORT RSAKeys, Strings, Base64, ASN1, BigNum;

FROM FileOps IMPORT
    (* const*)  NoSuchChannel,
    (* type *)  ChanId,
    (* proc *)  OpenOldFile, CloseFile, ReadLine;

FROM VarStrings IMPORT
    (* type *)  VarStringPtr, ByteStringPtr, ByteStr,
    (* proc *)  DiscardBS;

FROM BigNum IMPORT
    (* type *)  BN;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM LowLevel IMPORT
    (* proc *)  EVAL;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)
(*                          LOAD RSA PRIVATE KEY                        *)
(************************************************************************)

PROCEDURE LoadRSAPrivateKey (filename: ARRAY OF CHAR): RSAKeys.RSAKeyType;

    (* Gets RSA private key from a file.  *)
    (* This version handles the sort of file that starts with   *)
    (*  -----BEGIN PRIVATE KEY-----                             *)
    (* I'm not sure whether I need to handle othr formats.      *)

    CONST
        MaxChars = 32768;
        MaxBytes = 32768;

    VAR cid: ChanId;
        line: ARRAY [0..255] OF CHAR;
        pstring: VarStringPtr;
        pbytes: ByteStringPtr;
        part: ARRAY [0..2] OF ByteStr;
        keydata: ByteStr;
        numbers: ARRAY [0..9] OF BN;
        RSAKey: RSAKeys.RSAKeyType;

    BEGIN
        cid := OpenOldFile (filename, FALSE, FALSE);
        IF cid = NoSuchChannel THEN
            (* Fatal error *)
            WriteString ("FATAL ERROR: Cannot open ");
            WriteString (filename);  WriteLn;
            WriteString ("Program halted");  WriteLn;
            HALT;
        END (*IF*);

        (* Skip past the header line(s). *)

        REPEAT
            ReadLine (cid, line);
        UNTIL line[0] <> '-';

        (* Read the Base64 string. *)

        ALLOCATE (pstring, MaxChars);
        pstring^[0] := CHR(0);
        WHILE line[0] <> '-' DO
            Strings.Append (line, pstring^);
            ReadLine (cid, line);
        END (*WHILE*);
        CloseFile (cid);

        ALLOCATE (pbytes, MaxBytes);
        EVAL (Base64.Decode (pstring^, pbytes^));
        DEALLOCATE (pstring, MaxChars);

        (* Now we have the actual binary file.  The top level is a      *)
        (* sequence of three things: an integer, an object ID, and a    *)
        (* long octet string. We need only the octet string.            *)

        EVAL (ASN1.ExtractParts (pbytes^, part));
        keydata := ASN1.CopyContent (part[2].data^);

        EVAL (ASN1.GetNumericArray (keydata.data^, numbers));
        DEALLOCATE (pbytes, MaxBytes);
        DiscardBS (keydata);

        (* Assign the values to the fields of our key. *)

        BigNum.Discard (numbers[0]);
        RSAKey.n       := numbers[1];
        RSAKey.public  := numbers[2];
        RSAKey.private := numbers[3];
        RSAKey.p       := numbers[4];
        RSAKey.q       := numbers[5];
        RSAKey.dp      := numbers[6];
        RSAKey.dq      := numbers[7];
        RSAKey.qinv    := numbers[8];

        RETURN RSAKey;

    END LoadRSAPrivateKey;

(************************************************************************)

END TLSRSA.

