(**************************************************************************)
(*                                                                        *)
(*  Transport Layer Security                                              *)
(*  Copyright (C) 2024   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE TLSRecord;

        (********************************************************)
        (*                                                      *)
        (*                The TLS record protocol               *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            4 July 2023                     *)
        (*  Last edited:        12 November 2024                *)
        (*  Status:             Restart forced by major file    *)
        (*                      loss, but I think I've fixed    *)
        (*                      it all by now.                  *)
        (********************************************************)

(************************************************************************)

FROM SYSTEM IMPORT CARD8, CARD16, ADR;

IMPORT Strings;

FROM TLSBase IMPORT
    (* type *)  TextType,
    (* proc *)  NYI, DiscardTT;

FROM TLSCompression IMPORT
    (* type *)  CompressionMethod,
    (* proc *)  Compress, Decompress;

FROM TLSHMAC IMPORT
    (* const*)  MaxHMACSize,
    (* type *)  HMACType,
    (* proc *)  AppendHMACName, HMAClength, HMACkeylength, ComputeHMac;

FROM TLSCrypt IMPORT
    (* const*)  MaxBlockSize,
    (* type *)  BulkCipherAlgorithm, EncrRule,
    (* proc *)  InitEncryptionRule, SetEncryption, CloseEncryptionRule,
                Encrypt, Decrypt, AppendCipherName;

FROM Sockets IMPORT
    (* type *)  Socket,
    (* proc *)  send, recv;

FROM VarStrings IMPORT
    (* type *)  ByteStringPtr;

FROM Arith64 IMPORT
    (* type *)  CARD64LE,
    (* const*)  Zero64LE,
    (* proc *)  INC64LE;

FROM MiscFuncs IMPORT
    (* proc *)  Swap2, Swap4;

FROM LowLevel IMPORT
    (* proc *)  Copy, AddOffset;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransaction, LogTransactionL;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

TYPE
    Direction = (in, out);

    RLState = POINTER TO
                RECORD
                    server: BOOLEAN;
                    logID: TransactionLogID;
                    sock: Socket;
                    p: ARRAY Direction OF
                        RECORD
                            seqnum: CARD64LE;
                            compression: CompressionMethod;
                            mac: HMACType;
                            maclength: CARDINAL;
                            mackey: ARRAY [0..63] OF CARD8;
                            cipher: BulkCipherAlgorithm;
                            IV: ARRAY [0..MaxBlockSize-1] OF CARD8;
                            rule: EncrRule;
                        END (*RECORD*)
                END (*RECORD*);

(************************************************************************)
(*                            DEBUGGING STUFF                           *)
(************************************************************************)

PROCEDURE RecordTraceOn;

    (* For testing/debugging.  Logs data sent or received until         *)
    (* RecordTraceOff is called.  This is a module-wide setting rather  *)
    (* that being specific to one thread, so it can be used only for    *)
    (* simple tests.  (And should probably be deleted in the long run.) *)

    BEGIN
        (*NYI ("RecordTraceOn");*)
    END RecordTraceOn;

(************************************************************************)

PROCEDURE RecordTraceOff;

    (* Stops logging of data sent or received.  *)

    BEGIN
        (*NYI ("RecordTraceOff");*)
    END RecordTraceOff;

(************************************************************************)

PROCEDURE FirstNBytes (logID: TransactionLogID;  prefix: ARRAY OF CHAR;
                                        buffer: TextType;  N: CARDINAL);

    (* Writes out the first N bytes of buffer, including the header. *)

    BEGIN
        NYI (logID, "FirstNBytes");
    END FirstNBytes;

(************************************************************************)
(*                        KEYS AND MAC COMPUTATION                      *)
(************************************************************************)

PROCEDURE SetCipher (sess: RLState;  server_write: BOOLEAN;
                newcipher: BulkCipherAlgorithm;  newmac: HMACType;
                    MACKey: ARRAY OF CARD8;  MACKeyLength: CARDINAL;
                    cryptKey: ARRAY OF CARD8;  cryptKeyLength: CARDINAL;
                    newIV: ARRAY OF CARD8;  IVlength: CARDINAL);

    (* Sets cipher and keys for one direction.  *)

    TYPE labeltype = ARRAY Direction OF ARRAY [0..7] OF CHAR;

    CONST label = labeltype {"input", "output"};

    VAR d: Direction;
        message: ARRAY [0..79] OF CHAR;

    BEGIN
        IF server_write = sess^.server THEN d := out
        ELSE d := in;
        END (*IF*);
        WITH sess^.p[d] DO
            cipher := newcipher;
            mac := newmac;
            maclength := HMAClength (newmac);
            IF MACKeyLength > 0 THEN
                Copy (ADR(MACKey), ADR(mackey), MACKeyLength);
            END (*IF*);
            IF IVlength > 0 THEN
                Copy (ADR(newIV), ADR(IV), IVlength);
            END (*IF*);
            CloseEncryptionRule (rule);
            seqnum := Zero64LE;
            SetEncryption (rule, newcipher, cryptKey, cryptKeyLength);
        END (*WITH*);

        message := "Switched to ";
        Strings.Append (label[d], message);
        Strings.Append (" cipher ", message);
        AppendCipherName (newcipher, message);
        Strings.Append ("/", message);
        AppendHMACName (newmac, message);
        LogTransaction (sess^.logID, message);

    END SetCipher;

(************************************************************************)

PROCEDURE GenerateMAC (sess: RLState;  d: Direction;  buffer: TextType;
                            VAR (*OUT*) result: ARRAY OF CARD8);

    (* The input to the MAC computation is the concatenation of         *)
    (* sequence number, type, version, length, and buffer.fragment.     *)

    (* The input data to the MAC algorithm has a sequence       *)
    (* number prepended to buffer, and the HMAC algorithms do   *)
    (* not allow the data to be supplied in sections.  For now  *)
    (* the only solution I see is to make a copy of the data    *)
    (* to be processed.  I should check my HMACG to see         *)
    (* whether the steps can be broken up in a way that lets    *)
    (* the data be fed in in parts.                             *)

    VAR p, q: ByteStringPtr;
        input: TextType;
        N, N2, size, kl: CARDINAL;

    BEGIN
        IF sess^.p[d].mac <> nullhmac THEN

            (* Create a new buffer with the input to the MAC calculation.   *)

            input := buffer;
            size := Swap2 (buffer.length);
            N2 := size + 8 + 5;
            ALLOCATE (p, N2);

            (* Copy sequence number. *)

            N := Swap4 (sess^.p[d].seqnum.high);
            Copy (ADR(N), p, 4);
            q := AddOffset (p, 4);
            N := Swap4 (sess^.p[d].seqnum.low);
            Copy (ADR(N), q, 4);
            q := AddOffset (q, 4);

            (* Copy the five header bytes. *)

            Copy (ADR(buffer), q, 5);
            q := AddOffset (q, 5);

            (* Finally the data fragment. *)

            Copy (buffer.fragment, q, size);

            (* Now the MAC computation. *)

            kl := HMACkeylength (sess^.p[d].mac);
            ComputeHMac (sess^.p[d].mac, sess^.p[d].mackey, kl, p, N2, result);

            (* Discard the temporary data. *)

            DEALLOCATE (p, N2);
        END (*IF*);
    END GenerateMAC;

(************************************************************************)
(*                              OPEN/CLOSE                              *)
(************************************************************************)

PROCEDURE TLSRecordInit (S: Socket;  server: BOOLEAN;
                            logID: TransactionLogID): RLState;

    (* Initialises the record layer state for a TLS session. *)

    VAR sess: RLState;
        d: Direction;

    BEGIN
        NEW (sess);
        sess^.server := server;
        sess^.logID := logID;
        sess^.sock := S;
        FOR d := in TO out DO
            WITH sess^.p[d] DO
                seqnum := Zero64LE;
                compression := nocompression;
                mac := nullhmac;
                maclength := 0;
                cipher := nocipher;
                InitEncryptionRule(rule);
            END (*WITH*);
        END (*FOR*);
        RETURN sess;
    END TLSRecordInit;

(************************************************************************)

PROCEDURE TLSRecordClose (VAR (*INOUT*) sess: RLState);

    (* Closes this record layer state. *)

    BEGIN
        CloseEncryptionRule (sess^.p[in].rule);
        CloseEncryptionRule (sess^.p[out].rule);
        DISPOSE (sess);
    END TLSRecordClose;

(************************************************************************)

PROCEDURE DiscardFragment (VAR (*INOUT*) buffer: TextType);

    (* Releases memory held by this record. *)

    VAR N: CARDINAL;

    BEGIN
        IF buffer.length <> 0 THEN
            N := Swap2(buffer.length);
            DEALLOCATE (buffer.fragment, N);
        END (*IF*);
    END DiscardFragment;

(************************************************************************)
(*                                INPUT                                 *)
(************************************************************************)

PROCEDURE GetFragment (sess: RLState;  logID: TransactionLogID;
                                VAR (*OUT*) buffer: TextType): BOOLEAN;

    (* Receives one record from the sender, in raw form.  It is up to   *)
    (* the caller to work out what to do with this record.              *)

    (* NOTE: We store buffer.length in BigEndian format.    *)

    VAR mac, newmac: ARRAY [0..MaxHMACSize-1] OF CARD8;
        N, k: CARD16;
        match: BOOLEAN;

    BEGIN
        buffer.fragment := NIL;

        (* Get first 5 bytes of the input.  *)

        N := recv (sess^.sock, buffer, 5, 0);
        IF N <> 5 THEN
            LogTransactionL (sess^.logID, "Did not receive header");
            RETURN FALSE;
        END (*IF*);

        (* Now get the main part. *)

        N := Swap2(buffer.length);
        ALLOCATE (buffer.fragment, N);
        k := recv (sess^.sock, buffer.fragment^, N, 0);
        buffer.length := Swap2(N);
        IF k <> N THEN
            LogTransactionL (sess^.logID, "Did not receive fragment body");
            DEALLOCATE (buffer.fragment, N);
            buffer.length := 0;
            RETURN FALSE;
        END (*IF*);

        (* Do the decryption. *)

        Decrypt (sess^.p[in].rule, logID, buffer, mac, sess^.p[in].maclength);

        (* Check the mac. *)

        IF sess^.p[in].maclength > 0 THEN
            GenerateMAC (sess, in, buffer, newmac);
            match := TRUE;
            FOR k := 0 TO sess^.p[in].maclength-1 DO
                match := match AND (mac[k] = newmac[k]);
            END (*FOR*);
            IF NOT match THEN
                LogTransactionL (sess^.logID, "MAC mismatch in input");
            END (*IF*);
        END (*IF*);

        Decompress (sess^.p[in].compression, buffer);

        INC64LE (sess^.p[in].seqnum);

        RETURN TRUE;
    END GetFragment;

(************************************************************************)
(*                              OUTPUT                                  *)
(************************************************************************)

PROCEDURE PutFragment (sess: RLState;  logID: TransactionLogID;
                            VAR (*IN*) buffer: TextType);

    (* Sends buffer to the recipient.  This includes any necessary      *)
    (* encryption and compression.  The data in buffer^.fragment, a     *)
    (* TLSPlainText record, is then discarded.                          *)

    (* NOTE: The caller must use BigEndian format for buffer.length.    *)

    VAR mac: ARRAY [0..MaxHMACSize-1] OF CARD8;
        k, length: CARDINAL;

    BEGIN
        IF (buffer.majversion <> 3) OR (buffer.minversion <> 3) THEN
            LogTransactionL (logID, "version error in PutFragment");
            HALT;
        END (*IF*);

        (* First do compression and MAC calculation.  *)

        Compress (sess^.p[out].compression, buffer);
        GenerateMAC (sess, out, buffer, mac);

        (* Now the encryption.  This includes adding the MAC and padding. *)

        Encrypt (sess^.p[out].rule, logID, buffer, mac, sess^.p[out].maclength);

        (* Send buffer, including its header.*)

        k := send (sess^.sock, buffer, 5, 0);
        IF k <> 5 THEN
            LogTransactionL (sess^.logID, "Not all header output sent");
        END (*IF*);

        length := Swap2(buffer.length);
        k := send (sess^.sock, buffer.fragment^, length, 0);
        IF k <> length THEN
            LogTransactionL (sess^.logID, "Not all body output sent");
        END (*IF*);

        DiscardTT (buffer);

        INC64LE (sess^.p[out].seqnum);

    END PutFragment;

(************************************************************************)

END TLSRecord.

